import os

import numpy as np

from bird.meshing._mesh_tools import *
from bird.utilities.parser import parse_json, parse_yaml


def write_ofoam_preamble(outfile):
    outfile.write("FoamFile\n")
    outfile.write("{\n")
    outfile.write("    version     2.0;\n")
    outfile.write("    format      ascii;\n")
    outfile.write("    class       dictionary;\n")
    outfile.write("    object      blockMeshDict;\n")
    outfile.write("}\n")
    outfile.write("\n")
    outfile.write("convertToMeters 1.0;\n\n")


def write_vertices(outfile, lengths, nboxes):
    # nx,ny, nz are boxes
    delta = lengths[:] / nboxes[:]

    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("vertices\n(\n")
    for k in range(nboxes[2] + 1):
        for j in range(nboxes[1] + 1):
            for i in range(nboxes[0] + 1):
                outfile.write(
                    "( "
                    + str(i * delta[0])
                    + " "
                    + str(j * delta[1])
                    + "  "
                    + str(k * delta[2])
                    + ")\n"
                )
    outfile.write(");\n")


def write_this_block(outfile, comment, ids, mesh, zonename="none"):
    outfile.write("\n //" + comment + "\n")
    outfile.write("hex (")
    for i in range(len(ids)):
        outfile.write(str(ids[i]) + " ")
    outfile.write(")\n")

    if zonename != "none":
        outfile.write(zonename + "\n")

    outfile.write("( %d %d %d )\n" % (mesh[0], mesh[1], mesh[2]))
    outfile.write("SimpleGrading (1 1 1)\n")


def write_blocks(outfile, blockids, lengths, nboxes, points_per_len):
    nx = nboxes[0]
    ny = nboxes[1]
    nz = nboxes[2]

    idarray = np.zeros(8, dtype=int)
    mesharray = np.zeros(3, dtype=int)

    mesharray[:] = np.ceil(lengths[:] * points_per_len[:] / nboxes[:])

    nblocks = len(blockids)
    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("blocks\n(\n")

    for bl in range(nblocks):
        i = blockids[bl][0]
        j = blockids[bl][1]
        k = blockids[bl][2]

        for kk in range(2):
            for jj in range(2):
                for ii in range(2):
                    ri = ii
                    if jj == 1:
                        ri = 1 - ii  # reverse i
                    ind = (
                        (k + kk) * (nx + 1) * (ny + 1)
                        + (j + jj) * (nx + 1)
                        + (i + ri)
                    )
                    idarray[kk * 4 + jj * 2 + ii] = ind

        write_this_block(outfile, "block %d" % (bl), idarray, mesharray)

    outfile.write(");\n")


def write_patches(outfile):
    outfile.write(
        "\n// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n"
    )
    outfile.write("defaultPatch\n{   type wall;}\n\n")
    outfile.write("patches\n(\n")
    outfile.write(");\n")


def make_fluid_blocks_from_corner(corners):
    assert isinstance(corners, list)
    fluid_blocks = []
    for multi_block in corners:
        n_end_block = len(multi_block)
        for iend, end_block in enumerate(multi_block):
            # Add blocks that connect end points
            if iend == 0:
                # Make sure to not add endPoint if already there
                if list(end_block) not in fluid_blocks:
                    fluid_blocks.append(list(end_block))
            else:
                # Find direction to previous end point
                prev_block = multi_block[iend - 1]
                line = np.array(end_block, dtype=int) - np.array(
                    prev_block, dtype=int
                )
                nnz = np.nonzero(line)
                assert len(nnz) == 1
                direction = np.array(line / np.linalg.norm(line), dtype=int)
                blocks_line = []
                found_next = False
                count = 0
                # Add blocks until next end point is found
                while not found_next:
                    count += int(1)
                    new_block = (
                        np.array(prev_block, dtype=int) + count * direction
                    )
                    if (
                        np.linalg.norm(
                            new_block - np.array(end_block, dtype=int)
                        )
                        < 1
                    ):
                        found_next = True
                    blocks_line.append(
                        list(
                            np.array(prev_block, dtype=int) + count * direction
                        )
                    )
                fluid_blocks += blocks_line

    return fluid_blocks


def assemble_geom(input_file):
    # inpt = parse_json(input_file)
    if input_file.endswith(".yaml"):
        inpt = parse_yaml(input_file)
    elif input_file.endswith(".json"):
        inpt = parse_json(input_file)
    else:
        raise ValueError(f"unknown input file ({input_file}) extension")

    # ~~~~ Define background domain
    domain = inpt["Geometry"]["OverallDomain"]
    overall_nblocks = np.array(
        [
            domain["x"]["nblocks"],
            domain["y"]["nblocks"],
            domain["z"]["nblocks"],
        ],
        dtype=int,
    )
    overall_size_per_block = np.array(
        [
            domain["x"]["size_per_block"],
            domain["y"]["size_per_block"],
            domain["z"]["size_per_block"],
        ]
    )
    overall_size = overall_nblocks * overall_size_per_block

    # ~~~~ Define corners
    corners = [entry for entry in inpt["Geometry"]["Fluids"]]
    fluid_blocks = make_fluid_blocks_from_corner(corners)

    return {
        "overall_size": overall_size,
        "overall_nblocks": overall_nblocks,
        "fluid_blocks": fluid_blocks,
    }


def from_block_rect_to_seg(input_geom_dict, rescale=True):
    blocksize_x = input_geom_dict["OverallDomain"]["x"]["size_per_block"]
    if rescale:
        blocksize_x *= input_geom_dict["OverallDomain"]["x"]["rescale"]
    blocksize_y = input_geom_dict["OverallDomain"]["y"]["size_per_block"]
    if rescale:
        blocksize_y *= input_geom_dict["OverallDomain"]["y"]["rescale"]
    blocksize_z = input_geom_dict["OverallDomain"]["z"]["size_per_block"]
    if rescale:
        blocksize_z *= input_geom_dict["OverallDomain"]["z"]["rescale"]
    segments = {}
    iseg = 0
    for ifl, fluid_list in enumerate(input_geom_dict["Fluids"]):
        nblock = len(fluid_list)
        # if ifl > 0:
        #    segments[iseg] = {}
        #    segments[iseg]["blocks"] = [
        #        segments[iseg - 1]["blocks"][-1],
        #        fluid_list[i],
        #    ]
        #    iseg += 1
        for i in range(nblock - 1):
            segments[iseg] = {}
            segments[iseg]["blocks"] = [fluid_list[i], fluid_list[i + 1]]
            iseg += 1
    for iseg in segments:
        segments[iseg]["start"] = np.array(
            [
                blocksize_x * segments[iseg]["blocks"][0][0]
                + blocksize_x * 0.5,
                blocksize_y * segments[iseg]["blocks"][0][1]
                + blocksize_y * 0.5,
                blocksize_z * segments[iseg]["blocks"][0][2]
                + blocksize_z * 0.5,
            ]
        )
        segments[iseg]["end"] = np.array(
            [
                blocksize_x * segments[iseg]["blocks"][1][0]
                + blocksize_x * 0.5,
                blocksize_y * segments[iseg]["blocks"][1][1]
                + blocksize_y * 0.5,
                blocksize_z * segments[iseg]["blocks"][1][2]
                + blocksize_z * 0.5,
            ]
        )
        vec_conn = segments[iseg]["end"] - segments[iseg]["start"]
        segments[iseg]["conn"] = vec_conn
        norm_vec_conn = np.linalg.norm(vec_conn)
        segments[iseg]["normal_dir"] = int(np.nonzero(vec_conn)[0][0])
        if segments[iseg]["normal_dir"] == 0:
            segments[iseg]["max_rad"] = (blocksize_y + blocksize_z) / 4
        if segments[iseg]["normal_dir"] == 1:
            segments[iseg]["max_rad"] = (blocksize_x + blocksize_z) / 4
        if segments[iseg]["normal_dir"] == 2:
            segments[iseg]["max_rad"] = (blocksize_x + blocksize_y) / 4

    return {
        "segments": segments,
        "blocksize": [blocksize_x, blocksize_y, blocksize_z],
    }


def assemble_mesh(input_file):
    if input_file.endswith(".yaml"):
        inpt = parse_yaml(input_file)
    elif input_file.endswith(".json"):
        inpt = parse_json(input_file)
    else:
        raise ValueError(f"unknown input file ({input_file}) extension")

    # ~~~~ Define num points per block
    block_points = inpt["Meshing"]["Blockwise"]
    pperlen = np.array(
        [block_points["x"], block_points["y"], block_points["z"]], dtype=int
    )

    return {
        "pperlen": pperlen,
    }


def writeBlockMeshDict(out_folder, geomDict, meshDict):
    n_blocks = len(geomDict["fluid_blocks"])
    blockids = np.array(geomDict["fluid_blocks"]).reshape(n_blocks, 3)

    with open(os.path.join(out_folder, "blockMeshDict"), "w+") as outfile:
        write_ofoam_preamble(outfile)
        write_vertices(
            outfile, geomDict["overall_size"], geomDict["overall_nblocks"]
        )
        write_blocks(
            outfile,
            blockids,
            geomDict["overall_size"],
            geomDict["overall_nblocks"],
            meshDict["pperlen"],
        )
        write_patches(outfile)


def main(input_file, output_folder):
    geomDict = assemble_geom(input_file)
    meshDict = assemble_mesh(input_file)
    writeBlockMeshDict(output_folder, geomDict, meshDict)



import logging
import os

import corner
import jax
import jax.numpy as jnp
import jax.random as random
import numpy as np
from packaging import version

if version.parse(jax.__version__) >= version.parse("0.7.0"):
    # https://github.com/pyro-ppl/numpyro/issues/2051
    import jax.experimental.pjit
    from jax.extend.core.primitives import jit_p

    jax.experimental.pjit.pjit_p = jit_p

import numpyro
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS
from prettyPlot.plotting import *
from scipy.optimize import curve_fit

logger = logging.getLogger(__name__)


def plotAllEarly(data_dict, color_files=None, chop=False, extrap=False):
    fig = plt.figure()
    for idat, datf in enumerate(data_dict):
        if color_files is not None:
            color = color_files[idat]
        else:
            color = None
        if extrap:
            plt.plot(
                data_dict[datf]["textrap"],
                100 * data_dict[datf]["yextrap"],
                "--",
                color=color,
                label=f"{datf} y={100 * data_dict[datf]['yextrap'][-1]:.2f}",
            )
        if chop:
            lim = data_dict[datf]["lim"]
            plt.plot(
                data_dict[datf]["t"][lim:],
                100 * data_dict[datf]["y"][lim:],
                "o",
                color=color,
            )
        else:
            plt.plot(
                data_dict[datf]["t"], 100 * data_dict[datf]["y"], color=color
            )
    pretty_labels("time [s]", "yield [%]", 14)
    if extrap:
        pretty_legend(fontsize=14)


def plotAllEarly_uq(data_dict, color_files=None):
    fig = plt.figure()
    for idat, datf in enumerate(data_dict):
        color = color_files[idat]
        text = data_dict[datf]["textrap"]
        t = data_dict[datf]["t"]
        y = data_dict[datf]["y"] * 100
        med_real = data_dict[datf]["med_real"] * 100
        mean_real = data_dict[datf]["mean_real"] * 100
        std16_real = data_dict[datf]["std16_real"] * 100
        std84_real = data_dict[datf]["std84_real"] * 100

        plt.plot(text, med_real, color=color, linewidth=3)
        plt.fill_between(text, std16_real, std84_real, color=color, alpha=0.3)
        plt.plot(
            t,
            y,
            "o",
            color=color,
            markersize=5,
            linewidth=3,
            label=f"{datf} y={med_real[-1]:.2f}+/-{(std84_real[-1]-std16_real[-1])/2:.2f}",
        )
        pretty_labels("time [s]", "yield [%]", 14)
        pretty_legend(fontsize=14)


def smooth(y, box_pts):
    box = np.ones(box_pts) / box_pts
    y_smooth = np.convolve(y, box, mode="same")
    return y_smooth


def multi_data_load(data_root, tmax=600, data_files=None, color_files=None):
    if data_files is None:
        data_files = os.listdir(data_root)
    if color_files is None:
        color_files = [
            str(i * 0.75 / (len(data_files))) for i in range(len(data_files))
        ]
    data_dict = {}
    for idat, datf in enumerate(data_files):
        filename = os.path.join(data_root, datf)
        A = np.loadtxt(filename)
        data_dict[datf] = {}
        data_dict[datf]["t"] = A[:, 0]
        data_dict[datf]["y"] = A[:, 5] / np.clip(
            (A[:, 4] * 16 / 44 + A[:, 5]), a_min=1e-12, a_max=None
        )
        # chop data before increase and right after t=10s
        increase_ind_arr = np.argwhere(np.diff(data_dict[datf]["y"]) > 0)
        increase_ind = increase_ind_arr[
            np.argwhere(data_dict[datf]["t"][increase_ind_arr] > 10)[0][0]
        ][0]
        logger.info(
            f"data {datf} first time {data_dict[datf]['t'][increase_ind]:.2f}"
        )
        data_dict[datf]["lim"] = increase_ind
        y_fit = (
            data_dict[datf]["y"][increase_ind:]
            - data_dict[datf]["y"][increase_ind]
        )
        t_fit = (
            data_dict[datf]["t"][increase_ind:]
            - data_dict[datf]["t"][increase_ind]
        )
        data_dict[datf]["t_fit"] = t_fit
        data_dict[datf]["y_fit"] = y_fit
        data_dict[datf]["textrap"] = np.linspace(0, tmax, min(tmax, 1000))

    return data_dict, color_files


def sigm_fit(x, a, b, c, d, e, f):
    return 2 * c * (
        0.5 - 1 / (1 + np.exp(np.clip((a * x) ** b, a_min=None, a_max=20)))
    ) + 2 * f * (
        0.5 - 1 / (1 + np.exp(np.clip((d * x) ** e, a_min=None, a_max=20)))
    )


def fit_and_ext(
    data_dict,
    func=sigm_fit,
    bounds=([0, 0, 0, 0, 0, 0], [0.05, 1.7, np.inf, 0.015, 1.3, np.inf]),
):
    for idat, datf in enumerate(data_dict):
        popt, pcov = curve_fit(
            func,
            data_dict[datf]["t_fit"],
            data_dict[datf]["y_fit"],
            bounds=bounds,
        )
        data_dict[datf]["yextrap"] = func(data_dict[datf]["textrap"], *popt)
        logger.info(f"data {datf} coeff {popt}")
        lim_ind = data_dict[datf]["lim"]
        data_dict[datf]["textrap"] += data_dict[datf]["t"][lim_ind]
        data_dict[datf]["yextrap"] += data_dict[datf]["y"][lim_ind]

    return data_dict


def sigm_fit_jax(theta, x):
    a, b, M, c, d, N, sigma = theta
    F = (
        2
        * M
        * (0.5 - 1 / (1 + jnp.exp(jnp.clip((a * x) ** b, min=None, max=20))))
    )
    F += (
        2
        * N
        * (0.5 - 1 / (1 + jnp.exp(jnp.clip((c * x) ** d, min=None, max=20))))
    )
    return F


def bayes_step(x, y=None, y_err=0.1):
    # define parameters (incl. prior ranges)
    a = numpyro.sample("a", dist.Uniform(0, 0.05))
    b = numpyro.sample("b", dist.Uniform(0, 1.7))
    M = numpyro.sample("M", dist.Uniform(0, 1))
    c = numpyro.sample("c", dist.Uniform(0, 0.015))
    d = numpyro.sample("d", dist.Uniform(0.5, 1.3))
    N = numpyro.sample("N", dist.Uniform(0, 1))
    sigma = numpyro.sample("sigma", dist.Uniform(1e-2, 10))

    # implement the model
    # needs jax numpy for differentiability here
    y_model = sigm_fit_jax([a, b, M, c, d, N, sigma], x)

    # notice that we clamp the outcome of this sampling to the observation y
    numpyro.sample("obs", dist.Normal(y_model, sigma), obs=y)


def bayes_fit(data_dict, num_warmup=1000, num_samples=500):
    rng_key = random.PRNGKey(0)
    rng_key, rng_key_ = random.split(rng_key)
    for idat, datf in enumerate(data_dict):
        # Hamilton Markov Chain (HMC) with no u turn sampling (NUTS)
        kernel = NUTS(bayes_step, target_accept_prob=0.9)
        mcmc = MCMC(kernel, num_warmup=num_warmup, num_samples=num_samples)
        mcmc.run(
            rng_key_, x=data_dict[datf]["t_fit"], y=data_dict[datf]["y_fit"]
        )
        mcmc.print_summary()

        # Draw samples
        hmc_samples = mcmc.get_samples()
        labels = list(hmc_samples.keys())
        nsamples = len(hmc_samples[labels[0]])
        nparams = len(labels)
        np_hmc_samples = np.zeros((nsamples, nparams))
        for ilabel, label in enumerate(labels):
            if label == "a":
                nplabel = 0
            if label == "b":
                nplabel = 1
            if label == "M":
                nplabel = 2
            if label == "c":
                nplabel = 3
            if label == "d":
                nplabel = 4
            if label == "N":
                nplabel = 5
            if label == "sigma":
                nplabel = 6
            np_hmc_samples[:, nplabel] = np.array(hmc_samples[label])

        ## Post process
        # fig = corner.corner(np_hmc_samples, truths=theta, labels=labels)
        #
        #
        ## Convergence
        # fig, axes = plt.subplots(nparams, sharex=True)
        # for i in range(nparams):
        #    ax = axes[i]
        #    ax.plot(np_hmc_samples[:, i], "k", alpha=0.3, rasterized=True)
        #    ax.set_ylabel(labels[i])

        # data_dict[datf]["yextrap"] = f(data_dict[datf]["textrap"], *popt)
        # data_dict[datf]["textrap"] += data_dict[datf]["t"][increase_ind]
        # data_dict[datf]["yextrap"] += data_dict[datf]["y"][increase_ind]
        # Uncertainty propagation
        lim_ind = data_dict[datf]["lim"]
        nsamples = np_hmc_samples.shape[0]
        realization = []
        for i in range(nsamples):
            yext = sigm_fit_jax(
                np_hmc_samples[i, :], data_dict[datf]["textrap"]
            )
            yext += data_dict[datf]["y"][lim_ind]
            text = data_dict[datf]["textrap"] + data_dict[datf]["t"][lim_ind]
            if np.amax(yext) < 1:
                realization.append(yext)
        realization = np.array(realization)
        mean_real = np.mean(realization, axis=0)
        med_real = np.median(realization, axis=0)
        std84_real = np.percentile(realization, 84, axis=0)
        std16_real = np.percentile(realization, 16, axis=0)

        data_dict[datf]["textrap"] = text
        data_dict[datf]["mean_real"] = mean_real
        data_dict[datf]["med_real"] = med_real
        data_dict[datf]["std84_real"] = std84_real
        data_dict[datf]["std16_real"] = std16_real

    return data_dict




import numpy as np
from scipy.optimize import minimize

from bird import logger


def bounded_constraint(x):
    xend = 1 - np.sum(x)
    cons = 0
    for i in range(len(x)):
        cons += np.clip(-x[i], a_min=0, a_max=None)
        cons += np.clip((x[i] - 1), a_min=0, a_max=None)
    cons += np.clip(-xend, a_min=0, a_max=None)
    cons += np.clip((xend - 1), a_min=0, a_max=None)
    return 100 * cons


def std_constraint(x, meanTar, stdTar, diam):
    # x = np.clip(x, a_min=0, a_max=1)
    xend = 1 - np.sum(x)
    var = np.sum(x * (diam[:-1] - meanTar) ** 2)
    var += xend * (diam[-1] - meanTar) ** 2
    return 1e6 * abs(np.sqrt(np.clip(var, a_min=0, a_max=None)) - stdTar)


def mean_constraint(x, meanTar, diam):
    # x = np.clip(x, a_min=0, a_max=1)
    xend = 1 - np.sum(x)
    mean = np.sum(x * diam[:-1])
    mean += xend * diam[-1]
    return 1e6 * abs(mean - meanTar)


def bell_constraint(x, meanTar, diam):
    # x = np.clip(x, a_min=0, a_max=1)
    xend = 1 - np.sum(x)
    peakInd = np.argsort(abs(diam - meanTar))[0]
    cons = 0
    for i in range(peakInd):
        cons += np.clip(x[i] - x[i + 1], a_min=0, a_max=None)
    for i in range(peakInd + 1, len(x)):
        cons += np.clip(x[i] - x[i - 1], a_min=0, a_max=None)
    cons += np.clip(xend - x[-1], a_min=0, a_max=None)
    return 100 * cons


def objective(x, meanTar, stdTar, diam):
    return (
        bounded_constraint(x)
        + std_constraint(x, meanTar, stdTar, diam)
        + mean_constraint(x, meanTar, diam)
        + bell_constraint(x, meanTar, diam)
    )


def opt(meanTar, stdTar, diam):
    # res = minimize(objective, x0=0.33*np.ones(len(diam)-1), method="SLSQP")
    res = minimize(
        objective,
        args=(meanTar, stdTar, diam),
        x0=np.random.uniform(low=0, high=1, size=len(diam) - 1),
        method="SLSQP",
    )
    return res


def get_f_vals(meanTar, stdTar, diam):
    if meanTar < np.amin(diam) or meanTar > np.amax(diam):
        error_msg = f"mean target {meanTar} out of bounds [{np.amin(diam)}, {np.amax(diam)}]"
        logger.error(error_msg)
        raise ValueError(error_msg)
    tol = 10
    irep = 0
    while tol > 0.01:
        logger.debug(f"Rep = {irep}, tol={tol}")
        res = opt(meanTar, stdTar, diam)
        x = np.zeros(len(res.x) + 1)
        x[:-1] = res.x
        x[-1] = 1 - np.sum(res.x)
        x = np.clip(x, a_min=0, a_max=None)
        meanOut = np.sum(x * diam)
        stdOut = np.sqrt(
            np.sum(np.clip(x, a_min=0, a_max=None) * (diam - meanTar) ** 2)
        )
        tol = (
            abs(meanOut - meanTar) / meanTar
            + 0.1 * abs(stdOut - stdTar) / stdTar
        )
        irep += 1
        if irep > 100:
            logger.error(
                "optimization fail, typically occurs because the population balance domain is too tight"
            )

    logger.debug(f"meanTar = {meanTar}")
    logger.debug(f"stdTar = {stdTar}")
    logger.debug(f"x = {x}")
    logger.debug(f"mean = {np.sum(x * diam)}")
    logger.debug(
        f"std = {np.sqrt(np.sum(np.clip(x, a_min=0, a_max=None) * (diam - meanTar) ** 2))}"
    )
    return x


def poreDiamCorr(dp, ds, Ugs):
    g = 9.81
    sigmaL = 0.07
    rhoL = 1000
    muL = 8.90e-4
    Fr = Ugs**2 / (ds * g)
    We = rhoL * Ugs**2 * ds / sigmaL
    Re = rhoL * Ugs * ds / muL
    d32 = (
        ds
        * 7.35
        * (We ** (-1.7) * Re ** (0.1) * Fr ** (1.8) * (dp / ds) ** (1.7))
        ** 0.2
    )
    return d32, d32 * 0.15


if __name__ == "__main__":
    meanTar, stdTar = poreDiamCorr(2e-5, 0.15, 0.01)
    print(meanTar)
    print(stdTar)
    # meanTar = 2.1e-3
    # stdTar = meanTar*0.15
    diam = np.linspace(0.5e-3, 6e-3, 10)
    x = get_f_vals(meanTar, stdTar, diam)

    import matplotlib.pyplot as plt

    plt.plot(diam, x, "o")
    plt.show()



import argparse
import os
import time
from pathlib import Path

import joblib
import numpy as np
import tensorflow as tf
import tensorflow.keras as keras
from prettyPlot.progressBar import print_progress_bar
from sklearn.model_selection import train_test_split
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import MinMaxScaler, StandardScaler
from tensorflow.keras import initializers, layers, optimizers, regularizers

from bird import logger


def flexible_activation(x, activation):
    if activation is None:
        out = x
    elif activation.lower() == "swish":
        out = layers.Activation(swish_activation)(x)
    elif activation.lower() == "sigmoid":
        out = layers.Activation(activation="sigmoid")(x)
    elif activation.lower() == "softplus":
        out = layers.Activation(activation="softplus")(x)
    elif activation.lower() == "tanh":
        out = layers.Activation(activation="tanh")(x)
    elif activation.lower() == "elu":
        out = layers.Activation(activation="elu")(x)
    elif activation.lower() == "selu":
        out = layers.Activation(activation="selu")(x)
    elif activation.lower() == "gelu":
        out = layers.Activation(activation="gelu")(x)
    elif activation.lower() == "relu":
        out = layers.Activation(activation="relu")(x)
    elif activation.lower() == "leakyrelu":
        out = layers.LeakyReLU()(x)
    else:
        raise NotImplementedError(f"unknown activation {activation}")
    return out


def singleLayer(x, n_units, activation):
    out = layers.Dense(n_units)(x)
    out = flexible_activation(out, activation)
    return out


def makeModel(input_dim_par, output_dim, units, activation, final_activation):
    # inputs
    input_z = layers.Input(shape=(1,), name="input_z")
    input_par = layers.Input(shape=(input_dim_par,), name="input_par")
    input_z_par = layers.concatenate([input_z, input_par], name="input_z_par")
    x = input_z_par
    for unit in units:
        x = singleLayer(x, unit, activation)
    output = layers.Dense(output_dim)(x)
    output = flexible_activation(output, final_activation)
    model = keras.Model([input_z, input_par], output)
    return model


class Param_NN(keras.Model):
    def __init__(
        self,
        input_dim=None,
        output_dim=None,
        units=None,
        activation=None,
        final_activation=None,
        batch_size=None,
        nEpochs=None,
        weight_file=None,
        model_folder="Model",
        log_loss_folder="Log",
        np_data_file=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.np_data_file = np_data_file
        if input_dim is None or output_dim is None:
            input_dim, output_dim = self.get_dim_from_data(np_data_file)
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.input_dim_par = self.input_dim - 1

        self.model = makeModel(
            self.input_dim_par,
            self.output_dim,
            units,
            activation,
            final_activation,
        )
        if weight_file is not None:
            self.load(weight_file)

        self.batch_size = batch_size
        self.nEpochs = nEpochs

        self.model_folder = model_folder
        self.log_loss_folder = log_loss_folder

    def get_dim_from_data(self, data_file):
        tmp_dat = np.load(data_file)
        if len(tmp_dat["x"].shape) == 1:
            input_dim = 1
        else:
            input_dim = tmp_dat["x"].shape[1]
        if len(tmp_dat["y"].shape) == 1:
            output_dim = 1
        else:
            output_dim = tmp_dat["y"].shape[1]
        return input_dim, output_dim

    def make_data(self, np_file=None, scaler_folder="."):
        # Raw data
        if np_file is None:
            np_file = self.np_data_file

        tmp = np.load(np_file)
        z_dat = np.float32(np.reshape(tmp["x"][:, 0], (-1, 1)))
        par_dat = np.float32(tmp["x"][:, 1:])
        y_dat = np.float32(np.reshape(tmp["y"], (-1, 1)))

        # Scaler
        scaler_z = MinMaxScaler()
        scaler_par = MinMaxScaler()
        scaler_y = StandardScaler()
        scaler_z.fit(z_dat)
        scaler_par.fit(par_dat)
        scaler_y.fit(y_dat)
        joblib.dump(scaler_z, os.path.join(scaler_folder, "scaler_z.mod"))
        joblib.dump(scaler_par, os.path.join(scaler_folder, "scaler_par.mod"))
        joblib.dump(scaler_y, os.path.join(scaler_folder, "scaler_y.mod"))

        # Split and shuffle
        (
            z_train,
            z_test,
            par_train,
            par_test,
            y_train,
            y_test,
        ) = train_test_split(
            scaler_z.transform(z_dat),
            scaler_par.transform(par_dat),
            scaler_y.transform(y_dat),
            test_size=0.1,
            random_state=42,
        )

        return {
            "z_train": z_train,
            "par_train": par_train,
            "y_train": y_train,
            "z_test": z_test,
            "par_test": par_test,
            "y_test": y_test,
        }

    @tf.function
    def train_step(self, input_z=None, input_par=None, output_true=None):
        with tf.GradientTape() as tape:
            loss = self.calc_loss(input_z, input_par, output_true)
        grads = tape.gradient(loss, self.trainable_weights)
        self.optimizer.apply_gradients(zip(grads, self.trainable_weights))
        return {
            "loss": loss,
        }

    def calc_loss(self, input_z=None, input_par=None, output_true=None):
        output_predicted = self.model([input_z, input_par])
        loss = tf.math.reduce_sum(
            tf.math.square(output_predicted - output_true)
        )
        return loss

    def train(
        self,
        learningRateModel,
        batch_size=None,
        nEpochs=None,
        data_file=None,
        gradient_threshold=None,
    ):
        if gradient_threshold is not None:
            logger.info("clipping gradients at {gradient_threshold:.2g}")
        # Make sure the control file for learning rate is consistent with main.py, at least at first
        self.prepareLog()
        bestLoss = None

        if batch_size is None:
            batch_size = self.batch_size
        if nEpochs is None:
            nEpochs = self.nEpochs

        # Make dataset
        if data_file is None:
            data_file = self.np_data_file

        input_dim_data, output_dim_data = self.get_dim_from_data(data_file)
        assert input_dim_data == self.input_dim
        assert output_dim_data == self.output_dim

        dataset = self.make_data(
            scaler_folder=self.model_folder, np_file=data_file
        )
        self.n_batch = dataset["z_train"].shape[0] // batch_size
        self.freq = self.n_batch * 10
        # self.freq = 1

        train_dataset = tf.data.Dataset.from_tensor_slices(
            (
                dataset["z_train"],
                dataset["par_train"],
                dataset["y_train"],
            )
        ).batch(batch_size)
        test_dataset = tf.data.Dataset.from_tensor_slices(
            (
                dataset["z_test"],
                dataset["par_test"],
                dataset["y_test"],
            )
        ).batch(100000)

        # Prepare LR
        lr_m = learningRateModel
        self.optimizer = optimizers.Adam(learning_rate=lr_m)

        # Train
        print_progress_bar(
            0,
            nEpochs,
            prefix="Loss=%s  Epoch= %d / %d " % ("?", 0, nEpochs),
            suffix="Complete",
            length=20,
        )

        for epoch in range(nEpochs):
            train_loss = 0
            time_per_step = 0
            for step, batch in enumerate(train_dataset):
                self.total_step = step + epoch * self.n_batch
                time_s = time.time()
                loss_info = self.train_step(
                    input_z=batch[0], input_par=batch[1], output_true=batch[2]
                )
                loss_val = loss_info["loss"]
                time_e = time.time()
                train_loss = (
                    (step) * train_loss + tf.reduce_sum(loss_val)
                ) / (step + 1)
                time_per_step = (
                    (step) * (time_per_step) + (time_e - time_s)
                ) / (step + 1)
            test_loss = 0
            for val_step, val_batch in enumerate(test_dataset):
                val_loss = self.calc_loss(
                    val_batch[0], val_batch[1], val_batch[2]
                )
                test_loss = (
                    (val_step) * test_loss + tf.reduce_sum(val_loss)
                ) / (val_step + 1)

            print_progress_bar(
                epoch + 1,
                nEpochs,
                prefix=f"train_l={train_loss:.2f}, test_l={test_loss:.2f}, t/step={1e3*time_per_step:.2f} ms,  Epoch= {epoch + 1} / {nEpochs} ",
                suffix="Complete",
                length=20,
            )
            bestLoss = self.logTraining(
                epoch,
                loss=train_loss,
                val_loss=test_loss,
                bestLoss=bestLoss,
            )
            self.model.save_weights(
                os.path.join(self.model_folder, "last.weights.h5"),
                overwrite=True,
            )

    def pred(self, z, par, rescaled=False):
        if len(z.shape) == 1:
            z = np.reshape(z, (-1, 1))
        if len(par.shape) == 1:
            par = np.reshape(par, (-1, 1))
        scaler_z = joblib.load(os.path.join(self.model_folder, "scaler_z.mod"))
        scaler_par = joblib.load(
            os.path.join(self.model_folder, "scaler_par.mod")
        )
        scaler_y = joblib.load(os.path.join(self.model_folder, "scaler_y.mod"))
        z = np.float32(z)
        par = np.float32(par)
        if not rescaled:
            z_in = scaler_z.transform(z)
            par_in = scaler_par.transform(par)
        y_out = self.model([z_in, par_in])
        y = scaler_y.inverse_transform(y_out)
        return y

    def load(self, weight_file):
        self.model.load_weights(weight_file)

    def prepareLog(self):
        Path(self.model_folder).mkdir(parents=True, exist_ok=True)
        Path(self.log_loss_folder).mkdir(parents=True, exist_ok=True)
        try:
            os.remove(os.path.join(self.log_loss_folder, "log.csv"))
        except FileNotFoundError as err:
            print(err)
            pass
        # Make log headers
        f = open(os.path.join(self.log_loss_folder, "log.csv"), "a+")
        f.write("epoch;step;loss;recons_w_loss;recons_chi_loss\n")
        f.close()

    def logTraining(self, epoch, loss, val_loss, bestLoss):
        f = open(os.path.join(self.log_loss_folder, "log.csv"), "a+")
        f.write(
            str(int(epoch))
            + ";"
            + str(int(epoch * self.n_batch))
            + ";"
            + str(loss.numpy())
            + ";"
            + str(val_loss.numpy())
            + ";"
            + "\n"
        )
        f.close()
        if bestLoss is None or loss < bestLoss:
            bestLoss = loss
            self.model.save_weights(
                os.path.join(self.model_folder, "best.weights.h5"),
                overwrite=True,
            )
        return bestLoss

    def get_loss_dat(self):
        lossData = np.genfromtxt(
            os.path.join(self.log_loss_folder, "log.csv"),
            delimiter=";",
            skip_header=1,
        )
        return {
            "epoch": lossData[:, 0],
            "step": lossData[:, 1],
            "train_loss": lossData[:, 2],
            "val_loss": lossData[:, 3],
        }


if __name__ == "__main__":
    myNN = BCR_NN(
        input_dim=4,
        output_dim=1,
        units=[10, 20, 10, 5],
        activation="tanh",
        model_folder="Modeltmp",
        # weight_file='Modeltmp.save/last.weights.h5',
        log_loss_folder="Logtmp",
    )

    myNN.train(
        learningRateModel=1e-3,
        batch_size=32,
        nEpochs=1000,
        data_file="train_data_gh_17_raw.npz",
    )

    from plotsUtil import *

    nz = 32
    nz_true = 32
    A = np.load("train_data_gh_17_raw.npz")
    dat_par = np.repeat(np.reshape(A["x"][0, 1:], (1, -1)), nz, axis=0)
    dat_z = np.reshape(np.linspace(0, 4, nz), (-1, 1))
    pred = myNN.pred(dat_z, dat_par)

    fig = plt.figure()
    plt.plot(pred[:, 0], dat_z[:, 0], "o", label="pred")
    plt.plot(A["y"][:nz_true], A["x"][:nz_true, 0], "x", label="true")
    prettyLabels("gh", "z", 14)
    plotLegend()

    loss_dat = myNN.get_loss_dat()
    fig = plt.figure()
    plt.plot(
        loss_dat["epoch"],
        loss_dat["train_loss"],
        color="k",
        linewidth=3,
        label="train",
    )
    plt.plot(
        loss_dat["epoch"],
        loss_dat["val_loss"],
        color="b",
        linewidth=3,
        label="test",
    )
    prettyLabels("epoch", "loss", 14)
    ax = plt.gca()
    ax.set_yscale("log")
    plotLegend()

    plt.show()



import os
import re
from collections import defaultdict

import numpy as np
from ruamel.yaml import YAML

from bird import BIRD_CONST_DIR, logger
from bird.utilities.ofio import (
    get_species_name,
    read_openfoam_dict,
    write_openfoam_dict,
)
from bird.utilities.parser import parse_yaml


def get_species_properties(species_name: list[str]) -> dict:
    """
    Get thermo properties of species

    Parameters
    ----------
    species_name: list[str]
        List of species names

    Returns
    ----------
    species_prop: dict
        Dictionary of species properties
    """
    logger.debug(f"Reading properties of {species_name} from {BIRD_CONST_DIR}")
    species_prop = {}
    for spec_name in species_name:
        if spec_name == "water":
            species_file = "h2o.yaml"
        else:
            species_file = spec_name.lower() + ".yaml"
        species_dict = parse_yaml(os.path.join(BIRD_CONST_DIR, species_file))
        species_prop[spec_name] = species_dict
    return species_prop


def get_species_key_pair(
    thermo_properties: dict, species_name: list[str]
) -> dict:
    """
    Find the key where are stored the species info

    Parameters
    ----------
    thermo_properties: dict
        Dictionary of thermo properties
    species_name : list[str]
        List of species names

    Returns
    ----------
    pair_species_keys: dict
        Pair of species names and dictionary key
    """

    pair_species_keys = {}

    for spec_name in species_name:
        if spec_name in list(thermo_properties.keys()):
            pair_species_keys[spec_name] = spec_name
        else:
            found = False
            for key in list(thermo_properties.keys()):
                if spec_name in key:
                    pair_species_keys[spec_name] = key
                    logger.debug(f"Found '{spec_name}' in {key}")
                    found = True
                    break
            if not found:
                thermo_properties[spec_name] = {}
                pair_species_keys[spec_name] = spec_name
                logger.warning(
                    f"Could not find species '{spec_name}' info in {thermo_properties_file}"
                )

    return pair_species_keys


def compare_old_new_prop(
    key_seq_old: list[str],
    key_seq_new: list[str],
    old_dict: dict,
    new_dict: dict,
    spec_name: str,
    phase: str,
    type_name="float",
) -> str:
    """
    Compare old and new thermo properties and report any mismatch

    Parameters
    ----------
    key_seq_old: list[str]
        Sequence of keys in the old dict that gives access to the old property value
    key_seq_new: list[str]
        Sequence of keys in the new dict that gives access to the new property value
    old_dict: dict
        Old dictionary of values
    new_dict: dict
        New dictionary of values
    spec_name: str
        Name of the species (useful for reporting mismatch)
    phase: str
        Name of the phase (useful for reporting mismatch)
    type_name: str
        Name of the data type compared (useful for detecting mismatch).
        Can be 'float' or 'list' for now

    Returns
    ----------
    status: str
        Whether comparison was successful or not, allows for handling missing keys
    """

    assert type_name in ["float", "list"]

    status = ""
    try:
        old_val = old_dict
        for key in key_seq_old:
            old_val = old_val[key]
        if type_name == "float":
            old_val = float(old_val)
        elif type_name == "list":
            old_val = np.array([float(entry) for entry in old_val])
        new_val = new_dict
        for key in key_seq_new:
            new_val = new_val[key]
        if type_name == "float":
            new_val = float(new_val)
        elif type_name == "list":
            new_val = new_val.replace("(", "").replace(")", "").split()
            new_val = np.array([float(entry) for entry in new_val])
        if type_name == "float":
            if not abs(old_val - new_val) < 1e-12:
                logger.warning(
                    f"{key_seq_old[-1]} of '{spec_name}' ({phase}) updated from {old_val} to {new_val}"
                )
        elif type_name == "list":
            if not np.linalg.norm(old_val - new_val) < 1e-12:
                logger.warning(
                    f"{key_seq_old[-1]} of '{spec_name}' ({phase}) updated from {old_val} to {new_val}"
                )
        status = "success"
    except KeyError:
        status = "failure"
    return status


def update_gas_thermo_prop(
    thermo_properties: dict, species_prop: dict, pair_species_keys: dict
) -> dict:
    """
    Update gas thermo properties

    Parameters
    ----------
    thermo_properties: dict
        Dictionary of thermo properties
    species_prop : dict
        Dictionary of species properties
    pair_species_keys: dict
        Dictionary that maps species name to thermo properties key

    Returns
    ----------
    thermo_properties : dict
        Updated gas thermo properties
    """
    species_name = list(pair_species_keys.keys())

    for spec_name in species_name:
        key_val = pair_species_keys[spec_name]
        spec_dict = thermo_properties[key_val]
        # Molecular weight
        target_wm_dict = {
            "molWeight": str(species_prop[spec_name]["specie"]["molWeight"])
        }
        if not "specie" in spec_dict:
            spec_dict["specie"] = target_wm_dict
        else:
            status = compare_old_new_prop(
                key_seq_old=["specie", "molWeight"],
                key_seq_new=["molWeight"],
                old_dict=spec_dict,
                new_dict=target_wm_dict,
                spec_name=spec_name,
                phase="gas",
                type_name="float",
            )
            spec_dict["specie"]["molWeight"] = target_wm_dict["molWeight"]
            if status.lower() == "failure":
                spec_dict["specie"] = target_wm_dict

        # Thermo coeff
        thermo_keys = []
        thermo_keys_cp_coeff = []
        target_thermo_dict = {}
        if "Tlow" in species_prop[spec_name]["gas"]["thermodynamics"]:
            target_thermo_dict["Tlow"] = str(
                species_prop[spec_name]["gas"]["thermodynamics"]["Tlow"]
            )
            thermo_keys.append("Tlow")
        if "Thigh" in species_prop[spec_name]["gas"]["thermodynamics"]:
            target_thermo_dict["Thigh"] = str(
                species_prop[spec_name]["gas"]["thermodynamics"]["Thigh"]
            )
            thermo_keys.append("Thigh")
        if "Tcommon" in species_prop[spec_name]["gas"]["thermodynamics"]:
            target_thermo_dict["Tcommon"] = str(
                species_prop[spec_name]["gas"]["thermodynamics"]["Tcommon"]
            )
            thermo_keys.append("Tcommon")
        if "highCpCoeffs" in species_prop[spec_name]["gas"]["thermodynamics"]:
            target_thermo_dict["highCpCoeffs"] = (
                "( "
                + str(
                    species_prop[spec_name]["gas"]["thermodynamics"][
                        "highCpCoeffs"
                    ]
                )
                + " )"
            )
            thermo_keys_cp_coeff.append("highCpCoeffs")
        if "lowCpCoeffs" in species_prop[spec_name]["gas"]["thermodynamics"]:
            target_thermo_dict["lowCpCoeffs"] = (
                "( "
                + str(
                    species_prop[spec_name]["gas"]["thermodynamics"][
                        "lowCpCoeffs"
                    ]
                )
                + " )"
            )
            thermo_keys_cp_coeff.append("lowCpCoeffs")

        if not "thermodynamics" in spec_dict:
            spec_dict["thermodynamics"] = target_thermo_dict
        else:
            for thermo_key in thermo_keys:
                status = compare_old_new_prop(
                    key_seq_old=["thermodynamics", thermo_key],
                    key_seq_new=[thermo_key],
                    old_dict=spec_dict,
                    new_dict=target_thermo_dict,
                    spec_name=spec_name,
                    phase="gas",
                    type_name="float",
                )
                spec_dict["thermodynamics"][thermo_key] = target_thermo_dict[
                    thermo_key
                ]

            for thermo_key in thermo_keys_cp_coeff:
                status = compare_old_new_prop(
                    key_seq_old=["thermodynamics", thermo_key],
                    key_seq_new=[thermo_key],
                    old_dict=spec_dict,
                    new_dict=target_thermo_dict,
                    spec_name=spec_name,
                    phase="gas",
                    type_name="list",
                )
                spec_dict["thermodynamics"][thermo_key] = target_thermo_dict[
                    thermo_key
                ]

        # Transport coeff
        transport_keys = []
        target_transport_dict = {}
        if "As" in species_prop[spec_name]["gas"]["transport"]:
            target_transport_dict["As"] = str(
                species_prop[spec_name]["gas"]["transport"]["As"]
            )
            transport_keys.append("As")
        if "Ts" in species_prop[spec_name]["gas"]["transport"]:
            target_transport_dict["Ts"] = str(
                species_prop[spec_name]["gas"]["transport"]["Ts"]
            )
            transport_keys.append("Ts")
        if not "transport" in spec_dict:
            spec_dict["transport"] = target_transport_dict
        else:
            for trans_key in transport_keys:
                status = compare_old_new_prop(
                    key_seq_old=["transport", trans_key],
                    key_seq_new=[trans_key],
                    old_dict=spec_dict,
                    new_dict=target_transport_dict,
                    spec_name=spec_name,
                    phase="gas",
                    type_name="float",
                )
                spec_dict["transport"][trans_key] = target_transport_dict[
                    trans_key
                ]

        # Elements
        target_element_dict = species_prop[spec_name]["specie"]["elements"]
        if not "elements" in spec_dict:
            spec_dict["elements"] = target_element_dict

        thermo_properties[key_val] = spec_dict

    return thermo_properties


def update_liq_thermo_prop(
    thermo_properties: dict, species_prop: dict, pair_species_keys: dict
) -> dict:
    """
    Update liquid thermo properties

    Parameters
    ----------
    thermo_properties: dict
        Dictionary of thermo properties
    species_prop : dict
        Dictionary of species properties
    pair_species_keys: dict
        Dictionary that maps species name to thermo properties key

    Returns
    ----------
    thermo_properties : dict
        Updated liquid thermo properties
    """

    species_name = list(pair_species_keys.keys())
    for spec_name in species_name:
        key_val = pair_species_keys[spec_name]
        spec_dict = thermo_properties[key_val]

        # Molecular weight
        target_wm_dict = {
            "molWeight": str(species_prop[spec_name]["specie"]["molWeight"])
        }
        if not "specie" in spec_dict:
            spec_dict["specie"] = target_wm_dict
        else:
            status = compare_old_new_prop(
                key_seq_old=["specie", "molWeight"],
                key_seq_new=["molWeight"],
                old_dict=spec_dict,
                new_dict=target_wm_dict,
                spec_name=spec_name,
                phase="liquid",
                type_name="float",
            )
            spec_dict["specie"]["molWeight"] = target_wm_dict["molWeight"]
            if status.lower() == "failure":
                spec_dict["specie"] = target_wm_dict

        # Thermo coeff
        target_thermo_dict = {
            "Cv": "$CpMixLiq",
            "Hf": species_prop[spec_name]["liquid"]["thermodynamics"]["Hf"],
        }
        if not "thermodynamics" in spec_dict:
            spec_dict["thermodynamics"] = target_thermo_dict
        else:
            for thermo_key in ["Hf"]:
                status = compare_old_new_prop(
                    key_seq_old=["thermodynamics", thermo_key],
                    key_seq_new=[thermo_key],
                    old_dict=spec_dict,
                    new_dict=target_thermo_dict,
                    spec_name=spec_name,
                    phase="liquid",
                    type_name="float",
                )
                spec_dict["thermodynamics"][thermo_key] = target_thermo_dict[
                    thermo_key
                ]

        thermo_properties[key_val] = spec_dict

    return thermo_properties


def write_species_properties(case_folder: str, phase: str = "gas") -> None:
    """
    Write thermo properties open foam dict

    Parameters
    ----------
    case_folder: str
        Path to OpenFOAM case
    phase: str
        Name of phase where to find the species
    """
    logger.debug(
        f"Writing properties for phase '{phase}' in case {case_folder}"
    )
    species_name = get_species_name(case_folder=case_folder, phase=phase)
    species_prop = get_species_properties(species_name)
    thermo_properties_file = os.path.join(
        case_folder, "constant", f"thermophysicalProperties.{phase}"
    )
    thermo_properties = read_openfoam_dict(thermo_properties_file)
    pair_species_keys = get_species_key_pair(
        thermo_properties=thermo_properties, species_name=species_name
    )
    if phase == "gas":
        thermo_properties_update = update_gas_thermo_prop(
            thermo_properties, species_prop, pair_species_keys
        )
    if phase == "liquid":
        thermo_properties_update = update_liq_thermo_prop(
            thermo_properties, species_prop, pair_species_keys
        )
    filename = os.path.join(
        case_folder, "constant", f"thermophysicalProperties.{phase}"
    )
    write_openfoam_dict(thermo_properties_update, filename=filename)




import sys

import numpy as np

from bird import logger


def _check1D(signal: np.ndarray):
    """
    Check that array is 1D

    Parameters
    ----------
    signal: np.ndarray
        Signal data whose shape is checked
    """
    try:
        assert signal is not None
    except AssertionError:
        logger.error("Signal expected to be 1D but is None")
        raise ValueError
    try:
        assert np.sum(np.array(signal.shape) > 1) == 1
        signal = signal.flatten()
    except AssertionError:
        logger.error(f"Signal expected to be 1D, got shape {signal.shape}")
        raise ValueError


def _autocorr1D(time_series: np.ndarray) -> np.ndarray:
    """
    Compute normalized autocorrelation coefficient over different lag values

    Parameters
    ----------
    time_series: np.ndarray
        Time series of the signal to autocorrelated

    Returns
    ----------
    autocorr : np.ndarray
        Autocorrelation coefficient array over multiple lags
    """

    _check1D(time_series)
    time_series = time_series.flatten()
    time_series = time_series - np.mean(time_series)
    autocorr = np.correlate(time_series, time_series, mode="full")
    # Only keep the positive lags
    autocorr = autocorr[autocorr.size // 2 :]
    # Normalize
    autocorr /= autocorr[0]

    return autocorr


def _T0_fn(autocorr: np.ndarray) -> float:
    """
    Compute time between independent samples
    Following Trenberth, "Some Effects of Finite Sample Size and Persistence on Meteorological Statistics. Part I: Autocorrelations", 1984
    And Oliver et al., "Estimating uncertainties in statistics computed from direct numerical simulation", 2014

    Parameters
    ----------
    autocorr: np.ndarray
        Autocorrelation coefficient for different lag values

    Returns
    ----------
    T0 : float
        Time (in discretization step) for which samples can be considered independent.
    """
    _check1D(autocorr)
    autocorr = autocorr.flatten()

    N = len(autocorr)
    k = np.array(range(1, N))
    T0 = 1.0 + 2.0 * np.sum((1.0 - k / N) * autocorr[1:])

    # Check value of T0
    if T0 < 1.0 or T0 > N / 2:
        logger.warning(
            f"T0 value ({T0}) is suspicious, falling back to T0 = 1"
        )
        T0 = 1.0

    logger.debug(f"T0 = {T0}")

    return T0


def _to_equally_spaced(
    time_series: np.ndarray, time_values: np.ndarray | None
):
    """
    Transform to equally spaced time series

    Parameters
    ----------
    time_series: np.ndarray
        Time series of the signal
    time_values: np.ndarray|None
        The time values over which the time series is sampled.

    Returns
    ----------
    new_time_series, new_time_values: tuple
        Equally spaced version of time_series and time_values

    """
    if time_values is None:
        logger.debug(
            "Assuming time series sampled which equal spacing over time"
        )
        return time_series, time_values

    logger.debug("Making the time series equally spaced over time")
    _check1D(time_values)
    time_values = time_values.flatten()
    _check1D(time_series)
    time_series = time_series.flatten()

    min_time = time_values.min()
    max_time = time_values.max()
    dt = np.diff(time_values)
    min_diff = dt.min()
    max_diff = dt.max()
    try:
        assert min_diff > 0
    except AssertionError:
        logger.error(f"Time values are not ordered ({time_values})")
        raise ValueError

    # If already equally spaced, don't touch it
    if (dt.max() - dt.min()) < 1e-12 * (time_values.max() - time_values.min()):
        logger.debug("Time series already equally spaced")
        return time_series, time_values

    # Subsample, we could also oversample by dividing by min_diff
    n_val = int((max_time - min_time) / max_diff)
    new_time_values = np.linspace(min_time, max_time, n_val + 1)
    new_time_series = np.interp(new_time_values, time_values, time_series)
    logger.warning(
        f"New time series built with {n_val} pts instead of {len(time_series)} pts"
    )

    return new_time_series, new_time_values


def calc_mean(
    time_series: np.ndarray, time_values: np.ndarray | None = None
) -> tuple[float, float]:
    """
    Compute mean and the uncertainty about the mean, from a time-series

    Following Trenberth, "Some Effects of Finite Sample Size and Persistence on Meteorological Statistics. Part I: Autocorrelations", 1984
    And Oliver et al., "Estimating uncertainties in statistics computed from direct numerical simulation", 2014


    Parameters
    ----------
    time_series: np.ndarray
        Time series of the signal
    time_values: np.ndarray | None
        The time values over which the time series is sampled.
        If None, the time values are assumed equally spaced.
        Otherwise, time_values is used to create a new equally spaced time_values

    Returns
    ----------
    mean_val: float
        Mean value of the time_series
    unc_val: float
        95% uncertainty (1.96 sigma) about the mean

    """

    time_series, time_values = _to_equally_spaced(time_series, time_values)

    autocorr = _autocorr1D(time_series)
    N = len(time_series)
    T0 = _T0_fn(autocorr)

    mean_val = np.mean(time_series)
    sigsq = np.var(time_series) * N / (N - T0)
    unc_val = np.sqrt(sigsq * T0 / N)

    return mean_val, unc_val * 1.96



import os

import numpy as np
import stl

from bird import logger
from bird.preprocess.stl_patch.stl_shapes import *
from bird.utilities.parser import parse_json


def check_input(input_dict):
    assert isinstance(input_dict, dict)
    need_geom = False
    for bound in input_dict:
        if not bound == "Geometry" and not bound == "Meshing":
            assert isinstance(input_dict[bound], list)
            for patch in input_dict[bound]:
                assert isinstance(patch, dict)
                if "branch_id" in patch:
                    need_geom = True
    if need_geom:
        assert "Geometry" in input_dict
        assert "OverallDomain" in input_dict["Geometry"]
        assert "x" in input_dict["Geometry"]["OverallDomain"]
        assert "y" in input_dict["Geometry"]["OverallDomain"]
        assert "z" in input_dict["Geometry"]["OverallDomain"]
        assert "size_per_block" in input_dict["Geometry"]["OverallDomain"]["x"]
        assert "Fluids" in input_dict["Geometry"]
        assert isinstance(input_dict["Geometry"]["Fluids"], list)
        assert isinstance(input_dict["Geometry"]["Fluids"][0], list)


def get_all_vert_faces(input_dict, boundary_name):
    patch_mesh_list = create_boundary_patch_list(input_dict, boundary_name)
    boundary_mesh = STLMesh()
    boundary_mesh.from_mesh_list(patch_mesh_list)
    return boundary_mesh


def write_boundaries(input_dict, output_folder="."):
    check_input(input_dict)
    for boundary_name in input_dict.keys():
        if not boundary_name == "Geometry":
            logger.info(f"Making {boundary_name}")
            boundary_mesh = get_all_vert_faces(input_dict, boundary_name)
            logger.info(f"\tArea {boundary_mesh.area} m2")
            boundary_mesh.save(
                os.path.join(output_folder, f"{boundary_name}.stl")
            )



import numpy as np

from bird.meshing.block_rect_mesh import from_block_rect_to_seg
from bird.preprocess.dynamic_mixer.io_fvModels import *
from bird.preprocess.dynamic_mixer.mixer import Mixer
from bird.utilities.parser import parse_json


def check_input(input_dict):
    assert isinstance(input_dict, dict)
    mix_type = []
    for mix in input_dict["mixers"]:
        if "x" in mix:
            mix_type.append("expl")
        else:
            mix_type.append("loop")
    if "loop" in mix_type:
        assert "Geometry" in input_dict
        assert "OverallDomain" in input_dict["Geometry"]
        assert "x" in input_dict["Geometry"]["OverallDomain"]
        assert "y" in input_dict["Geometry"]["OverallDomain"]
        assert "z" in input_dict["Geometry"]["OverallDomain"]
        assert "size_per_block" in input_dict["Geometry"]["OverallDomain"]["x"]
        assert "Fluids" in input_dict["Geometry"]
        assert isinstance(input_dict["Geometry"]["Fluids"], list)
        assert isinstance(input_dict["Geometry"]["Fluids"][0], list)

    return mix_type


def write_fvModel(input_dict, output_folder=".", force_sign=False):
    mix_type = check_input(input_dict)
    write_preamble(output_folder)
    if "loop" in mix_type:
        geom_dict = from_block_rect_to_seg(input_dict["Geometry"])
        mesh_dict = input_dict["Meshing"]
    for imix, mtype in enumerate(mix_type):
        mixer = Mixer()
        if mtype == "expl":
            mixer.update_from_expl_dict(input_dict["mixers"][imix])
            if mixer.ready:
                if force_sign:
                    write_mixer_force_sign(mixer, output_folder)
                else:
                    write_mixer(mixer, output_folder)
        elif mtype == "loop":
            mixer.update_from_loop_dict(
                input_dict["mixers"][imix], geom_dict, mesh_dict
            )
            if mixer.ready:
                if force_sign:
                    write_mixer_force_sign(mixer, output_folder)
                else:
                    write_mixer(mixer, output_folder)

    write_end(output_folder)




import math
import os
import re

import numpy as np

from bird import logger


def _check_phase_name(phase: str):
    """
    Check that phase name is valid

    Parameters
    ----------
    phase: str
        Name of phase where to find the species
    """
    try:
        assert phase in ["gas", "liquid"]
    except AssertionError:
        error_msg = f"Phase name ('{phase}') is not in ['gas', 'liquid']"
        logger.error(error_msg)
        raise NotImplementedError(error_msg)


def _read_mesh(filename: str) -> np.ndarray:
    """
    Reads cell center location from meshCellCentres_X.obj

    Parameters
    ----------
    filename: str
        meshCellCentres_X.obj filename

    Returns
    -------
    cell_centers: np.ndarray
        Array (N,3) representing the cell centers (N is number of cells)
    """

    assert "meshCellCentres" in filename
    assert ".obj" in filename
    cell_centers = np.loadtxt(filename, usecols=(1, 2, 3))
    return cell_centers


def _ofvec2arr(vec: str) -> np.ndarray:
    """
    Converts a vector written as a string into a numpy array

    Parameters
    ----------
    vec: str
        Vector written as string
        Must start with "("
        Must end with ")"

    Returns
    -------
    vec_array: np.ndarray
        Array (3,) representing the vector
    """

    vec = vec.strip()
    assert vec[0] == "("
    assert vec[-1] == ")"

    vec_list = vec[1:-1].split()
    vec_array = np.array([float(entry) for entry in vec_list])
    return vec_array


def _is_comment(line: str) -> bool:
    """
    Checks if line is a comment

    Parameters
    ----------
    line: str
        Line of file

    Returns
    -------
    is_comment: bool
        True if line is a comment
        False if line is not a comment
    """
    is_comment = False

    sline = line.strip()
    if sline.startswith("//"):
        is_comment = True
    elif sline.startswith("/*"):
        is_comment = True
    return is_comment


def _read_meta_data(filename: str, mode: str | None = None) -> dict:
    """
    Read meta data from field outputted by OpenFOAM in ASCII format

    Parameters
    ----------
    filename: str
        Field filename
    mode: str | None
        If "scalar", expects a scalar field
        If "vector", expects a vector field
        If None, obtained from field header

    Returns
    -------
    meta_data: dict
        Dictionary that contain info about the scalar field
            ============= =====================================================
            Key           Description (type)
            ============= =====================================================
            name          Name of the field (*str*)
            n_cells       Number of computational cells (*int*)
            uniform       Whether the field is uniform (*bool*)
            uniform_value Uniform value if uniform field (*float* | *np.ndarray*)
            type          "vector" or "scalar" (*str*)
            ============= =====================================================
    """
    meta_data = {}
    meta_data["type"] = mode

    # Read meta data
    with open(filename, "r") as f:
        header_done = False
        iline = 0
        while not header_done:
            line = f.readline()
            if not _is_comment(line):
                # Get field type
                if (line.strip().startswith("class")) and (";" in line):
                    sline = line.strip().split()
                    field_type = sline[1][:-1]
                    if field_type.lower() == "volvectorfield":
                        field_type = "vector"
                    elif field_type.lower() == "volscalarfield":
                        field_type = "scalar"
                    else:
                        err_msg = f"Field type {field_type} not recognized"
                        err_msg = "Only 'volVectorField' and 'volScalarField' are supported"
                        raise NotImplementedError(err_msg)
                    if mode is not None:
                        assert field_type == mode
                    meta_data["type"] = field_type
                # Get field name
                if (line.strip().startswith("object")) and (";" in line):
                    sline = line.strip().split()
                    field_name = sline[1][:-1]
                    meta_data["name"] = field_name

                # Check if uniform
                if line.strip().startswith("internalField"):
                    if "nonuniform" in line:
                        meta_data["uniform"] = False
                        iline += 1
                        # read until no comments
                        comment = True
                        while comment:
                            count_line = f.readline().strip()
                            if not _is_comment(count_line):
                                comment = False
                        try:
                            n_cells = int(count_line)
                            meta_data["n_cells"] = n_cells
                            header_done = True
                        except ValueError:
                            raise ValueError(
                                f"Expected integer number of cells on line {iline}, got: '{count_line}'"
                            )
                    elif "uniform" in line:
                        meta_data["uniform"] = True
                        if meta_data["type"].lower() == "scalar":
                            sline = line.split()
                            unif_value = float(sline[-1].strip(";"))
                        elif meta_data["type"].lower() == "vector":
                            sline = line.split()
                            for ientry, entry in enumerate(sline):
                                if ";" in entry:
                                    ind_end = ientry
                                    break
                            line_cropped = " ".join(sline[2 : ientry + 1])
                            unif_value = _ofvec2arr(line_cropped.strip(";"))
                        else:
                            raise NotImplementedError(
                                f"Mode {mode} is unknown (must be 'scalar', 'vector' or None)"
                            )
                        meta_data["uniform_value"] = unif_value
                        header_done = True

            if len(line) == 0 and (not header_done):
                raise ValueError(
                    f"File {filename} ends before meta-data found"
                )

    return meta_data


def _readOFScal(
    filename: str,
    n_cells: int | None = None,
    n_header: int | None = None,
    meta_data: dict | None = None,
) -> dict:
    """
    Read a scalar field outputted by OpenFOAM in ASCII format

    Parameters
    ----------
    filename: str
        Field filename
    n_cells : int | None
        Number of computational cells in the domain
    n_header : int | None
        Number of header lines
    meta_data : dict | None
        meta data dictionary
        If None, it is read from filename

    Returns
    -------
    data: dict
        Dictionary that contain info about the scalar field
            ======== =====================================================
            Key      Description (type)
            ======== =====================================================
            field    For nonuniform fields, array of size the number of cells (*np.ndarray*).
                     For uniform fields with a specified n_cells,
                     array of size the number of cells (*np.ndarray*).
                     For uniform fields, a scalar value (*float*)
            name     Name of the scalar field (*str*)
            n_cells  Number of computational cells (*int*)
            n_header Number of header lines (*int*)
            ======== =====================================================
    """
    field = None

    if meta_data is None:
        # Read meta data
        meta_data = _read_meta_data(filename, mode="scalar")

    # Set field
    if meta_data["uniform"]:
        if n_cells is None:
            field = meta_data["uniform_value"]
        else:
            field = meta_data["uniform_value"] * np.ones(n_cells)
    else:
        n_cells = meta_data["n_cells"]

        # Get header size
        if n_header is None:
            oldline = ""
            newline = ""
            iline = 0
            eof = False
            with open(filename, "r") as f:
                while iline < 100 and (not eof):
                    line = f.readline()
                    if len(line) == 0:
                        eof = True
                    oldline = newline
                    newline = line
                    if str(n_cells) in oldline and "(" in newline:
                        n_header = iline + 1
                        break
                    iline += 1
                else:
                    raise ValueError(
                        "Could not find a sequence {n_cells} and '(' in file ({filename})"
                    )

        # Rapid field read
        try:
            field = np.loadtxt(filename, skiprows=n_header, max_rows=n_cells)
        except Exception as err:
            error_msg = f"Issue when reading {filename}"
            logger.error(error_msg)
            raise err(error_msg)

    return {
        "field": field,
        "name": meta_data["name"],
        "n_cells": n_cells,
        "n_header": n_header,
    }


def _readOFVec(
    filename: str,
    n_cells: int | None = None,
    n_header: int | None = None,
    meta_data: dict | None = None,
) -> dict:
    """
    Read a vector field outputted by OpenFOAM in ASCII format

    Parameters
    ----------
    filename: str
        Vector field filename
    n_cells : int | None
        Number of computational cells in the domain
    n_header : int | None
        Number of header lines
    meta_data : dict | None
        meta data dictionary
        If None, it is read from filename

    Returns
    -------
    data: dict
        Dictionary that contain info about the scalar field
            ======== =====================================================
            Key      Description (type)
            ======== =====================================================
            field    For nonuniform fields, array of size the number of cells by 3 (*np.ndarray*).
                     For uniform fields with a specified n_cells,
                     array of size the number of cells by 3 (*np.ndarray*).
                     For uniform fields, a scalar value (*float*)
            name     Name of the field (*str*)
            n_cells  Number of computational cells (*int*)
            n_header Number of header lines (*int*)
            ======== =====================================================
    """
    field = None

    if meta_data is None:
        # Read meta data
        meta_data = _read_meta_data(filename, mode="vector")

    # Set field
    if meta_data["uniform"]:
        if n_cells is None:
            field = meta_data["uniform_value"]
        else:
            field = meta_data["uniform_value"] * np.ones((n_cells, 3))
    else:
        n_cells = meta_data["n_cells"]
        if n_header is None:
            oldline = ""
            newline = ""
            iline = 0
            eof = False
            with open(filename, "r") as f:
                while iline < 100 and (not eof):
                    line = f.readline()
                    if len(line) == 0:
                        eof = True
                    oldline = newline
                    newline = line
                    if str(n_cells) in oldline and "(" in newline:
                        n_header = iline + 1
                        break
                    iline += 1
                else:
                    raise ValueError(
                        "Could not find a sequence {n_cells} and '(' in file ({filename})"
                    )

        try:
            field = np.loadtxt(
                filename, dtype=tuple, skiprows=n_header, max_rows=n_cells
            )
            for i in range(n_cells):
                field[i, 0] = float(field[i, 0][1:])
                field[i, 1] = float(field[i, 1])
                field[i, 2] = float(field[i, 2][:-1])
            field = np.array(field).astype(float)

        except Exception as err:
            error_msg = f"Issue when reading {filename}"
            logger.error(error_msg)
            raise err(error_msg)

    return {
        "field": field,
        "name": meta_data["name"],
        "n_cells": n_cells,
        "n_header": n_header,
    }


def _readOF(
    filename: str,
    n_cells: int | None = None,
    n_header: int | None = None,
    meta_data: dict | None = None,
) -> dict:
    """
    Read an OpenFOAM field outputted in ASCII format

    Parameters
    ----------
    filename: str
        Field filename
    n_cells : int | None
        Number of computational cells in the domain
    n_header : int | None
        Number of header lines
    meta_data : dict | None
        meta data dictionary
        If None, it is read from filename


    Returns
    -------
    data: dict
        Dictionary that contain info about the scalar field
            ======== =====================================================
            Key      Description (type)
            ======== =====================================================
            field    For nonuniform fields, array of size the number of cells (*np.ndarray*).
                     For uniform fields with a specified n_cells,
                     array of size the number of cells (*np.ndarray*).
                     For uniform fields, a scalar value (*float*)
            name     Name of the field (*str*)
            n_cells  Number of computational cells (*int*)
            n_header Number of header lines (*int*)
            ======== =====================================================
    """
    if meta_data is None:
        # Read meta data
        meta_data = _read_meta_data(filename)
    if meta_data["type"] == "scalar":
        return _readOFScal(filename=filename, meta_data=meta_data)
    if meta_data["type"] == "vector":
        return _readOFVec(filename=filename, meta_data=meta_data)


def read_field(
    case_folder: str,
    time_folder: str,
    field_name: str,
    n_cells: int | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Read field at a given time and store it in dictionary for later reuse

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    field_name: str
        Name of the field file to read
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    field : np.ndarray | float
        Field read
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    if not (field_name in field_dict) or field_dict[field_name] is None:
        # Read field if it had not been read before
        field_file = os.path.join(case_folder, time_folder, field_name)
        field = _readOF(field_file, n_cells=n_cells)["field"]
        field_dict[field_name] = field
    else:
        # Get field from dict if it has been read before
        field = field_dict[field_name]

    return field, field_dict


def read_size_groups(case_folder: str) -> dict:
    """
    Get the bubble size groups represented by the number density fields (fX.gas)

    Parameters
    ----------
    case_folder: str
        Path to case folder

    Returns
    ----------
    ndf_groups: dict
        Dictionary describing the number density fields
        Key is the name of the number density field (fX)
        Value is a dictionary with keys 'diam' and 'bin_size'
        corresponding to the bubble diameter and the bin size in m
    """

    phaseProperties_file = os.path.join(
        case_folder, "constant", "phaseProperties"
    )
    phaseProperties = read_openfoam_dict(phaseProperties_file)

    # Make sure that population balance is used
    try:
        assert phaseProperties["populationBalances"] == ["bubbles"]
        assert phaseProperties["gas"]["diameterModel"] == "velocityGroup"
        assert (
            phaseProperties["gas"]["velocityGroupCoeffs"]["populationBalance"]
            == "bubbles"
        )
    except AssertionError:
        logger.warning(
            "Reading size groups for a case where population balance is not used"
        )

    size_grouptmp = phaseProperties["gas"]["velocityGroupCoeffs"]["sizeGroups"]
    logger.debug(f"Found {len(size_grouptmp)} number density fields")

    # Associate number density field to size
    size_group = {}
    for name in size_grouptmp:
        size_group[name] = float(size_grouptmp[name]["dSph"])

    # Sort by size in ascending order
    size_group = dict(sorted(size_group.items(), key=lambda item: item[1]))

    # Get the bin size
    bin_size_group = {}
    group_names = list(size_group.keys())
    logger.warning(
        "The bin sizes definition make assumption of uniformity and need to be revisited if used"
    )
    for igroup, name in enumerate(group_names):
        if igroup == 0:
            bin_size = (
                size_group[group_names[igroup + 1]]
                - size_group[group_names[igroup]]
            )
        elif igroup == len(group_names) - 1:
            bin_size = (
                size_group[group_names[igroup]]
                - size_group[group_names[igroup - 1]]
            )
        else:
            bin_size_p = (
                size_group[group_names[igroup + 1]]
                - size_group[group_names[igroup]]
            )
            bin_size_m = (
                size_group[group_names[igroup]]
                - size_group[group_names[igroup - 1]]
            )
            assert abs(bin_size_p - bin_size_m) < 1e-12
            bin_size = bin_size_m
        bin_size_group[name] = bin_size

    # Put size and bin size together
    ndf_groups = {}
    for name in group_names:
        ndf_groups[name] = {
            "diam": size_group[name],
            "bin_size": bin_size_group[name],
        }

    return ndf_groups


def get_case_times(
    case_folder: str, remove_zero: bool = False
) -> tuple[list[float], list[str]]:
    """
    Get list of all time folders from an OpenFOAM case

    Parameters
    ----------
    case_folder: str
        Path to case folder
    remove_zero : bool
        Whether to remove zero from the time folder list

    Returns
    -------
    time_float_sorted: list[float]
        List of time folder values in ascending order
    time_str_sorted: list[str]
        List of time folder names in ascending order

    """
    # Read Time
    times_tmp = os.listdir(case_folder)
    # remove non floats
    for i, entry in reversed(list(enumerate(times_tmp))):
        try:
            a = float(entry)
            if remove_zero:
                if abs(a) < 1e-12:
                    _ = times_tmp.pop(i)
        except ValueError:
            logger.debug(f"{entry} not a time folder, removing")
            a = times_tmp.pop(i)
            # print('removed ', a)
    time_float = [float(entry) for entry in times_tmp]
    time_str = [entry for entry in times_tmp]
    index_sort = np.argsort(time_float)
    time_float_sorted = [time_float[i] for i in list(index_sort)]
    time_str_sorted = [time_str[i] for i in list(index_sort)]

    return time_float_sorted, time_str_sorted


def _get_mesh_time(case_folder: str) -> str | None:
    """
    Get the time at which the mesh was printed

    Parameters
    ----------
    case_folder: str
        Path to case folder

    Returns
    ----------
    time_mesh: str | None
        The name of the time at which "meshCellCentresXXX" was created
        If None, nothing was found
    """

    files_tmp = os.listdir(case_folder)
    time_mesh = None
    for entry in files_tmp:
        if "meshCellCentres" in entry:
            time_mesh = entry[16:-4]

    return time_mesh


def _get_volume_time(case_folder: str) -> str | None:
    """
    Get the time at which the volume was printed

    Parameters
    ----------
    case_folder: str
        Path to case folder

    Returns
    ----------
    time_volume: str | None
        The name of the time at which "V" was created
        If None, nothing was found
    """

    time_float, time_str = get_case_times(case_folder)
    time_volume = None
    for entry in time_str:
        if os.path.exists(os.path.join(case_folder, entry, "V")):
            logger.debug(f"Volume time found to be {entry}")
            time_volume = entry
            break

    return time_volume


def _remove_comments(text: str) -> str:
    """
    Remove C++-style comments (// and /*) from the input and markers like #{ #}

    Parameters
    ----------
    text: str
        Raw input text containing comments

    Returns
    ----------
    text: str
        Text with all comments removed
    """

    # text = re.sub(
    #    r"/\*.*?\*/", "", text, flags=re.DOTALL
    # )  # Remove /* */ comments
    # text_unc = re.sub(r"//.*", "", text)  # Remove // comments

    text = re.sub(r"/\*.*?\*/", "", text, flags=re.DOTALL)
    text = re.sub(r"//.*", "", text)
    text = re.sub(r"#\{", "{", text)
    text = re.sub(r"#\};", "}", text)
    text = re.sub(r"#codeStream", "", text)
    return text


def _tokenize(text: str) -> list[str]:
    """
    Add spaces around special characters (brace and semicolons) to make them separate tokens

    Parameters
    ----------
    text: str
        The cleaned (comment-free) OpenFOAM-style text.

    Returns
    ----------
    token_list: list[str]
        List of tokens.
    """
    text = re.sub(r"([{}();])", r" \1 ", text)
    text = re.sub(r'"\s*\(\s*([^)]+?)\s*\)\s*"', r'"(\1)"', text)

    token_list = text.split()

    # print(token_list)
    # print(text)

    return token_list


def _parse_tokens(tokens: list[str]) -> dict:
    """
    Parse OpenFOAM tokens into a nested Python dictionary.
    Special handling for `code { ... }` blocks to be stored as raw strings.

    Parameters
    ----------
    tokens: list[str]
        A list of tokens produced by `_tokenize`.

    Returns
    ----------
    parsed: dict
        A nested dictionary that represents the OpenFOAM dictionary.

    """

    def parse_block(index: int) -> tuple:
        result = {}
        while index < len(tokens):
            token = tokens[index]
            if token == "}":
                return result, index + 1
            elif token == "{":
                raise SyntaxError("Unexpected '{'")

            key = token
            index += 1

            # key followed by dictionary
            if index < len(tokens) and tokens[index] == "{":
                index += 1
                if key == "code":
                    code_lines = []
                    while tokens[index] != "}":
                        code_lines.append(tokens[index])
                        index += 1
                    index += 1
                    if index < len(tokens) and tokens[index] == ";":
                        index += 1
                    result[key] = " ".join(code_lines).strip()
                else:
                    subdict, index = parse_block(index)
                    result[key] = subdict

            # key followed by list
            elif index < len(tokens) and tokens[index] == "(":
                index += 1

                # Peek to check if it's a dict-list (starts with '(' then '{')
                if tokens[index] == "(":
                    dictlist = {}
                    while tokens[index] != ")":
                        if tokens[index] != "(":
                            raise SyntaxError(
                                f"Expected '(' for label in dict-list, got {tokens[index]}"
                            )
                        # Read full label (e.g., "(gas and liquid)")
                        label_tokens = []
                        while tokens[index] != ")":
                            label_tokens.append(tokens[index])
                            index += 1
                        label_tokens.append(tokens[index])  # include ')'
                        index += 1
                        label = " ".join(label_tokens)

                        if tokens[index] != "{":
                            raise SyntaxError(
                                f"Expected '{{' after label {label}"
                            )
                        index += 1
                        subdict, index = parse_block(index)
                        dictlist[label] = subdict
                    index += 1  # skip final ')'
                    if index < len(tokens) and tokens[index] == ";":
                        index += 1
                    result[key] = dictlist
                else:
                    # Standard list or dict-like list (e.g. sizeGroups)
                    lst = []
                    dictlist = {}

                    while tokens[index] != ")":
                        label = tokens[index]
                        index += 1

                        if index < len(tokens) and tokens[index] == "{":
                            # Inline dict entry like: f1 { dSph 1e-3; value 0.0; }
                            index += 1
                            subdict, index = parse_block(index)
                            dictlist[label] = subdict
                        else:
                            # Skip semicolons if present (e.g. in lists like species)
                            if label != ";":
                                lst.append(label)

                    index += 1  # skip ')'
                    if index < len(tokens) and tokens[index] == ";":
                        index += 1

                    # Choose dictlist only if it has content; otherwise, use lst
                    result[key] = dictlist if dictlist else lst

            # key followed by scalar
            elif index < len(tokens):
                value = tokens[index]
                index += 1
                if index < len(tokens) and tokens[index] == ";":
                    index += 1
                result[key] = value

        return result, index

    parsed, _ = parse_block(0)
    return parsed


def read_openfoam_dict(filename: str) -> dict:
    """
    Parse OpenFOAM dictionary into a python dictionary

    Parameters
    ----------
    filename: str
        OpenFOAM dictionary filename

    Returns
    -------
    dict_of: dict
        A Python dictionary representing the structure of the OpenFOAM dictionary.
    """
    with open(filename, "r+") as f:
        text = f.read()
    text = _remove_comments(text)
    tokens = _tokenize(text)
    foam_dict = _parse_tokens(tokens)
    return foam_dict


def write_openfoam_dict(data: dict, filename: str, indent: int = 0) -> None:
    """
    Save a Python dictionary back to an OpenFOAM-style file.

    Parameters
    ----------
    d: dict
        Python dictionary to save
    filename: str
        The file that will contain the saved dictionary
    indent: int
        Number of indentation space
    """

    def write_block(f, key, value, indent=0):
        pad = " " * indent
        if isinstance(value, dict):
            f.write(f"{pad}{key}\n{pad}{{\n")
            for k, v in value.items():
                write_block(f, k, v, indent + 4)
            f.write(f"{pad}}}\n")
        elif isinstance(value, list):
            if all(isinstance(v, str) for v in value):
                f.write(f"{pad}{key}\n{pad}(\n")
                for v in value:
                    f.write(f"{pad}    {v}\n")
                f.write(f"{pad});\n")
            else:
                # assume list of numbers for OpenFOAM vectors
                joined = " ".join(value)
                f.write(f"{pad}{key}    ( {joined} );\n")
        else:
            f.write(f"{pad}{key}    {value};\n")

    with open(filename, "w") as f:
        # Write OpenFOAM header
        f.write(
            r"""/*--------------------------------*- C++ -*----------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Version:  9
     \\/     M anipulation  |
\*---------------------------------------------------------------------------*/
"""
        )
        # Write FoamFile block first
        foam_file = data.pop("FoamFile", None)
        if foam_file:
            write_block(f, "FoamFile", foam_file)
        f.write(
            "// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //\n\n"
        )

        # Then write the rest of the blocks
        for key, value in data.items():
            write_block(f, key, value)
            f.write("\n")

        # Write OpenFOAM footer
        f.write(
            "// ************************************************************************* //\n"
        )


def read_cell_centers(
    case_folder: str,
    cell_centers_file: str | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray, dict]:
    """
    Read field of cell centers and store it in dictionary for later reuse

    Parameters
    ----------
    case_folder: str
        Path to case folder
    cell_centers_file : str
        Filename of cell center data
        If None, find the cell center file automoatically
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    cell_centers : np.ndarray
        cell centers read from file
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    if (
        not ("cell_centers" in field_dict)
        or field_dict["cell_centers"] is None
    ):
        if cell_centers_file is None:
            # try to find the mesh time
            mesh_time = _get_mesh_time(case_folder)
            if mesh_time is not None:
                cell_centers_file = f"meshCellCentres_{mesh_time}.obj"

        try:
            cell_centers = _read_mesh(
                os.path.join(case_folder, cell_centers_file)
            )
            field_dict["cell_centers"] = cell_centers

        except FileNotFoundError:

            error_msg = f"Could not find {cell_centers_file}"
            error_msg += "You can generate it with\n\t"
            error_msg += f"`writeMeshObj -case {case_folder}`\n"
            time_float, time_str = get_case_times(case_folder)
            correct_path = f"meshCellCentres_{time_str[0]}.obj"
            if not correct_path == cell_centers_file:
                error_msg += (
                    f"And adjust the cell center file path to {correct_path}"
                )
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)
    else:
        cell_centers = field_dict["cell_centers"]

    return cell_centers, field_dict


def read_cell_volumes(
    case_folder: str,
    time_folder: str | None = None,
    n_cells: int | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Read volume at a given time and store it in dictionary for later reuse

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str | None
        Name of time folder to analyze.
        If None, it will be found automatically
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    cell_volumes : np.ndarray | float
        Field of cell volumes
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }

    if not ("V" in field_dict) or field_dict["V"] is None:
        if time_folder is None:
            # Find the time at which the volume was printed
            time_folder = _get_volume_time(case_folder)
            kwargs_vol["time_folder"] = time_folder
        try:
            cell_volumes, field_dict = read_field(
                field_name="V", field_dict=field_dict, **kwargs_vol
            )

        except FileNotFoundError:
            error_msg = f"Could not find {os.path.join(case_folder, time_folder, 'V')}\n"
            time_float, time_str = get_case_times(case_folder)
            error_msg += "You can generate V with\n\t"
            error_msg += f"`postProcess -func writeCellVolumes -time {time_str[0]} -case {case_folder}`"
            logger.error(error_msg)
            raise FileNotFoundError(error_msg)
    else:
        # Get field from dict if it has been read before
        cell_volumes = field_dict["V"]

    return cell_volumes, field_dict


def read_bubble_diameter(
    case_folder: str,
    time_folder: str | None = None,
    n_cells: int | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Read bubble diameter at a given time and store it in dictionary for later reuse.
    A specific function is constructed so that if d.gas is not available, the bubble
    diameter is read from phaseProperties.

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str | None
        Name of time folder to analyze.
        If None, it will be found automatically
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    cell_volumes : np.ndarray | float
        Field of cell volumes
    field_dict : dict
        Dictionary of fields read
    """

    if field_dict is None:
        field_dict = {}

    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }

    if not ("d.gas" in field_dict) or field_dict["d.gas"] is None:
        try:
            d_gas, field_dict = read_field(
                field_name="d.gas", field_dict=field_dict, **kwargs
            )

        except FileNotFoundError as err:
            logger.debug(
                "Could not find d.gas, checking if bubble size model is constant"
            )
            # d.gas does not exist, it might be because a constant bubble diameter model is used
            phaseProperties_dict = read_openfoam_dict(
                os.path.join(case_folder, "constant", "phaseProperties")
            )
            # If the bubble size model is not constant, raise original error
            if (
                not phaseProperties_dict["gas"]["diameterModel"].lower()
                == "constant"
            ):
                logger.error(
                    f"Bubble size model is not constant ({phaseProperties_dict['gas']['diameterModel']}), yet could not find d.gas"
                )
                raise FileNotFoundError(err)
            else:
                # Get the constant bubble diameter
                logger.debug("Reading bubble diameter from phaseProperties")
                d_gas = float(
                    phaseProperties_dict["gas"]["constantCoeffs"]["d"]
                )
                field_dict["d.gas"] = d_gas
    else:
        # Get field from dict if it has been read before
        d_gas = field_dict["d.gas"]

    return d_gas, field_dict


def get_species_name(case_folder: str, phase: str = "gas") -> list[str]:
    """
    Get list of species name in a phase

    Parameters
    ----------
    case_folder: str
        Path to OpenFOAM case
    phase: str
        Name of phase where to find the species

    Returns
    ----------
    species_name: list[str]
        List of species name in the phase
    """
    _check_phase_name(phase)
    logger.debug(f"Finding species in phase '{phase}'")

    thermo_properties = read_openfoam_dict(
        os.path.join(
            case_folder, "constant", f"thermophysicalProperties.{phase}"
        )
    )

    try:
        species = thermo_properties["species"]
        if not isinstance(species, list):
            assert isinstance(species, str)
            species = [species]
    except KeyError:
        species = []
    try:
        defaultSpecie = thermo_properties["defaultSpecie"]
        if not isinstance(defaultSpecie, list):
            assert isinstance(defaultSpecie, str)
            defaultSpecie = [defaultSpecie]
    except KeyError:
        defaultSpecie = []
    try:
        inertSpecie = thermo_properties["inertSpecie"]
        if not isinstance(inertSpecie, list):
            assert isinstance(inertSpecie, str)
            inertSpecie = [inertSpecie]
    except KeyError:
        inertSpecie = []

    species_name = list(set(species + defaultSpecie + inertSpecie))
    logger.debug(f"Species in phase '{phase}' are {species_name}")
    return species_name


def read_mu_liquid(
    case_folder: str,
    time_folder: str | None = None,
    n_cells: int | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Read liquid viscosity at a given time and store it in dictionary for later reuse.
    A specific function is constructed so that if thermo:mu.liquid is not available, the liquid viscosity is read from globalVars

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str | None
        Name of time folder to analyze.
        If None, it will be found automatically
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    cell_volumes : np.ndarray | float
        Field of cell volumes
    field_dict : dict
        Dictionary of fields read
    """

    if field_dict is None:
        field_dict = {}

    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }

    if (
        not ("thermo:mu.liquid" in field_dict)
        or field_dict["thermo:mu.liquid"] is None
    ):
        try:
            mu_liq, field_dict = read_field(
                field_name="thermo:mu.liquid", field_dict=field_dict, **kwargs
            )

        except FileNotFoundError as err:
            logger.debug(
                "Could not find thermo:mu.liquid, checking if it can be read from globalVars"
            )
            # thermo:mu.liquid does not exist
            mu_liq = None
            globalVars = read_global_vars(case_folder, cross_ref=True)
            thermo = read_openfoam_dict(
                os.path.join(
                    case_folder, "constant", "thermophysicalProperties.liquid"
                )
            )
            liquid_species = get_species_name(case_folder, phase="liquid")
            main_liq_species = None
            for species_name in liquid_species:
                if "water" in liquid_species:
                    main_liq_species = "water"
                    break
                if "WATER" in liquid_species:
                    main_liq_species = "WATER"
                    break
                if "h2o" in liquid_species:
                    main_liq_species = "h2o"
                    break
                if "H2O" in liquid_species:
                    main_liq_species = "H2O"
                    break
            for key in thermo:
                if main_liq_species in key:
                    liq_species_prop = thermo[key]
                    break
            if liq_species_prop["transport"]["mu"] == "$muMixLiq":
                # You are using a constant mu
                mu_liq = globalVars["muMixLiq"]
            if mu_liq is None:
                logger.error(
                    f"Liquid viscosity is not constant, yet could not find thermo:mu.liquid"
                )
                raise FileNotFoundError(err)
            else:
                # Get the constant bubble diameter
                logger.debug("Reading liquid viscosity from globalVars")
                field_dict["thermo:mu.liquid"] = mu_liq

    else:
        # Get field from dict if it has been read before
        mu_liq = field_dict["thermo:mu.liquid"]

    return mu_liq, field_dict


def _cross_reference_global_vars(
    globalVars_dict: dict,
) -> dict:
    """
    Resolve cross referencing in the globalVars dictionary read
    so that all the values are float
    """

    cross_referenced_globalVars_dict = globalVars_dict.copy()

    # capture $Variable references
    var_pattern = re.compile(r"\$([A-Za-z0-9_]+)")

    for key, value in globalVars_dict.items():
        # Check if we need to do cross referencing
        if isinstance(value, str) and value.startswith("#calc"):

            # Extract expression inside quotes
            expr = value.split('"', 1)[1].rsplit('"', 1)[0]

            # Replace $Var with its numeric value from globalVars_dict
            def repl(match):
                varname = match.group(1)
                if varname not in cross_referenced_globalVars_dict:
                    raise KeyError(
                        f"Variable '{varname}' not found for expression {expr}"
                    )
                return str(cross_referenced_globalVars_dict[varname])

            expr = var_pattern.sub(repl, expr)

            # Replace 'Foam::pow' with 'math.pow'
            if "Foam::pow" in expr:
                expr = expr.replace("Foam::pow", "math.pow")

            if "pow" in expr:
                indstr = [m.start() for m in re.finditer("pow", expr)]
                new_expr = expr
                count_replace = 0
                for ind in indstr:
                    if not "mat" in new_expr[ind - 6 : ind + 3]:
                        tmp = new_expr
                        new_expr = (
                            tmp[: ind + 5 * count_replace]
                            + "math.pow"
                            + tmp[ind + 5 * count_replace + 3 :]
                        )
                        count_replace += 1
                expr = new_expr
            # Replace 'exp' with 'math.exp'
            expr = expr.replace("exp", "math.exp")

            try:
                result = eval(expr, {"math": math})
            except Exception as e:
                raise RuntimeError(
                    f"Error evaluating globalVars expression for {key}: {expr}"
                ) from e

            # Convert to int if whole number
            if isinstance(result, float) and result.is_integer():
                result = int(result)

            cross_referenced_globalVars_dict[key] = result

    return cross_referenced_globalVars_dict


def read_global_vars(
    case_folder: str | None = None,
    filename: str | None = None,
    cross_ref: bool = True,
) -> dict:
    """
    Read globalVars into a python dictionary

    Parameters
    ----------
    case_folder: str | None
        Path to case folder
        If None, using filename
    filename: str | None
        Path to the globalVars file
        If None, using case_folder
    cross_ref: bool
        Do cross referencing or no
        If True, all the `#calc` values from the python dict are replaced by their numeral values
        If False, all the `#calc` values are stored as string

    Returns
    ----------
    globalVars_dict : dict
        Dictionary that contains the globals vars variable values
    """

    # Set the filename to read
    if case_folder is None:
        if filename is None or not os.path.isfile(filename):
            error_msg = f"Global Vars file ({filename}) not found, but is needed if case folder not specified"
            raise FileNotFoundError(error_msg)
    else:
        if not os.path.exists(case_folder):
            error_msg = f"Case folder ({case_folder}) not found, but is needed if global vars filename not specified"
            raise FileNotFoundError(error_msg)
        filename = os.path.join(case_folder, "constant", "globalVars")

    logger.debug(f"Reading {filename}")

    # Now read globalVars
    globalVars_dict = {}
    with open(filename, "r") as f:
        for line in f:
            # Remove comments
            line = line.split("//", 1)[0].strip()
            if not line:
                continue  # skip empty/comment-only lines

            # Ensure it ends with ;
            if not line.endswith(";"):
                continue
            line = line[:-1].strip()  # remove trailing ;

            # Split key and value
            parts = line.split(None, 1)  # split on first whitespace
            if len(parts) != 2:
                continue
            key, value = parts
            logger.debug(f"\tReading {key}")

            # Keep calc expressions as-is
            if value.startswith("#calc"):
                globalVars_dict[key] = value
            else:
                # Try numeric conversion otherwise store value as str
                try:
                    val = float(value)
                    if val.is_integer():
                        val = int(val)
                    globalVars_dict[key] = val
                except ValueError:
                    globalVars_dict[key] = value
    if cross_ref:
        # Remove the #calc stuff
        globalVars_dict = _cross_reference_global_vars(globalVars_dict)

    return globalVars_dict


def species_name_to_mw(case_folder: str, species_name: str) -> float:
    r"""
    Get molecular weight in :math:`kg/mol` from the species name.
    In order of availability, the molecular weight is read from the thermophysicalProperties.XXX,
    and globalVars.
    If nothing is found in globalVars (last resort) an error is raised.
    This function is primarily useful to compute concentrations

    Parameters
    ----------
    case_folder: str
        Path to case folder
    species_name: str
        The name of the species for which molecular weight is desired

    Returns
    ----------
    mw : float
        The species molecular weight
    """
    thermo_gas_filename = os.path.join(
        case_folder, "constant", "thermophysicalProperties.gas"
    )
    thermo_liq_filename = os.path.join(
        case_folder, "constant", "thermophysicalProperties.liquid"
    )
    globalVars_filename = os.path.join(case_folder, "constant", "globalVars")

    # Handle corner case of len 1 list of strings, and interpret it as string
    if (
        isinstance(species_name, list)
        and len(species_name) == 1
        and isinstance(species_name[0], str)
    ):
        species_name = species_name[0]
    assert isinstance(species_name, str)

    # Special case for water
    if species_name.lower() in ["water", "h2o"]:
        species_names = ["water", "h2o", "H2O", "WATER"]
    else:
        species_names = [
            species_name,
            species_name.upper(),
            species_name.lower(),
        ]

    mw = None
    # Try finding the molecular weight from thermo liquid
    if os.path.isfile(thermo_liq_filename):
        thermo = read_openfoam_dict(thermo_liq_filename)
        for name in species_names:
            if name in thermo:
                mw = float(thermo[name]["specie"]["molWeight"]) * 1e-3
                break
        # Handle situation of type ("species1|species2") for the key
        if mw is None:
            for name in species_names:
                for key in thermo:
                    if name in key:
                        try:
                            mw = (
                                float(thermo[name]["specie"]["molWeight"])
                                * 1e-3
                            )
                            break
                        except:
                            pass
        if mw is not None:
            logger.debug(
                f"Read the {species_name} molecular weight ({mw}) from thermophysicalProperties.liquid"
            )
        else:
            logger.debug(
                f"Could not read the {species_name} molecular weight from thermophysicalProperties.liquid"
            )

    # Try finding the molecular weight from thermo gas
    if mw is None and os.path.isfile(thermo_gas_filename):
        thermo = read_openfoam_dict(thermo_gas_filename)
        for name in species_names:
            if name in thermo:
                mw = float(thermo[name]["specie"]["molWeight"]) * 1e-3
                break
        # Handle situation of type ("species1|species2") for the key
        if mw is None:
            for name in species_names:
                for key in thermo:
                    if name in key:
                        try:
                            mw = (
                                float(thermo[name]["specie"]["molWeight"])
                                * 1e-3
                            )
                            break
                        except:
                            pass
        if mw is not None:
            logger.debug(
                f"Read the {species_name} molecular weight ({mw}) from thermophysicalProperties.gas"
            )
        else:
            logger.debug(
                f"Could not read the {species_name} molecular weight from thermophysicalProperties.gas"
            )

    # Last resort: try finding the molecular weight from thermo gas
    if mw is None and os.path.isfile(globalVars_filename):
        globalVars = read_global_vars(case_folder=case_folder, cross_ref=True)
        for name in species_names:
            if f"Mw_{name}" in globalVars:
                mw = float(globalVars[f"Mw_{name}"])
                break
        if mw is not None:
            mw = globalVars[f"Mw_{name}"]
        else:
            err_msg = f"Mw_{species_name} was not found in globalVars."
            err_msg += "\nIf you add it, it should be [kg/mol]"
            raise KeyError(err_msg)

    return mw



import numpy as np
import vtk.numpy_interface.dataset_adapter as dsa
from paraview import simple as pv

from bird import logger
from bird.utilities.ofio import *

from .kla_utils import compute_kla


def _field_filter(
    field: float | np.ndarray, ind: np.ndarray, field_type: str
) -> float | np.ndarray:
    """
    Filter field by index. Handle uniform and non uniform fields

    Parameters
    ----------
    field: float | np.ndarray
        Field to filter
    ind: np.ndarray
        Cell indices to keep
    field_type : str
        Type of the field ("scalar" or "vector")

    Returns
    ----------
    filtered_field: float | np.ndarray
        Field filtered by cell indices

    """
    if field_type.lower() == "scalar":
        if isinstance(field, np.ndarray):
            if len(field.shape) > 1:
                err_msg = f"Scalar field shape {field.shape} but expected a flat array"
                raise ValueError(err_msg)
            filtered_field = field[ind]
        elif isinstance(field, float):
            # Uniform field
            filtered_field = field
        else:
            err_msg = f"Got field type {type(field)}."
            err_msg += " Expected float or np.ndarray for scalar field"
            raise TypeError(err_msg)

    elif field_type.lower() == "vector":
        if isinstance(field, np.ndarray):
            if field.shape == (3,):
                # Uniform field
                filtered_field = field
            else:
                filtered_field = field[ind]
        else:
            err_msg = f"Got field type {type(field)}."
            err_msg += " Expected np.ndarray for vector field"
            raise TypeError(err_msg)

    else:
        msg = f"Field type ({field_type}) not recognized"
        msg += " Supported field types are 'scalar' and 'vector'"
        raise NotImplementedError(msg)

    return filtered_field


def _get_ind_liq(
    case_folder: str,
    time_folder: str,
    threshold: float = 0.5,
    n_cells: int | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Get indices of pure liquid cells (where alpha.liquid > threshold)
    Threshold is 0.5 by default

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    threshold: float
        Liquid is when alpha_liq > threshold
        Assumes threshold = 0.5 by default
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    field_name: str
        Name of the field file to read
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    ind_liq : np.ndarray | float
        indices of pure liquid cells
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        return None

    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    assert threshold <= 1
    assert threshold >= 0

    logger.warning(
        f"Assuming that alpha_liq > {threshold} denotes pure liquid"
    )

    # Compute indices of pure liquid
    if not ("ind_liq" in field_dict) or field_dict["ind_liq"] is None:
        alpha_liq, field_dict = read_field(
            case_folder,
            time_folder,
            field_name="alpha.liquid",
            n_cells=n_cells,
            field_dict=field_dict,
        )
        ind_liq = np.argwhere(alpha_liq > threshold)[:, 0]
        field_dict["ind_liq"] = ind_liq
    else:
        ind_liq = field_dict["ind_liq"]

    return ind_liq, field_dict


def _get_ind_gas(
    case_folder: str,
    time_folder: str,
    threshold: float = 0.5,
    n_cells: int | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Get indices of pure gas cells (where alpha.liquid <= threshold)
    Threshold is 0.5 by default

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    threshold: float
        Gas is when alpha_liq <= threshold
        Assumes threshold = 0.5 by default
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    field_name: str
        Name of the field file to read
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    ind_gas : np.ndarray | float
        indices of pure gas cells
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    assert threshold <= 1
    assert threshold >= 0

    logger.warning(f"Assuming that alpha_liq <= {threshold} denotes pure gas")

    # Compute indices of pure liquid
    if not ("ind_gas" in field_dict) or field_dict["ind_gas"] is None:
        alpha_liq = read_field(
            case_folder,
            time_folder,
            field_name="alpha.liquid",
            n_cells=n_cells,
            field_dict=field_dict,
        )
        ind_gas = np.argwhere(alpha_liq <= threshold)
        field_dict["ind_gas"] = ind_gas
    else:
        ind_gas = field_dict["ind_gas"]

    return ind_gas, field_dict


def _get_ind_slice(
    case_folder: str,
    location: float,
    direction: int | None = None,
    tolerance: float | None = None,
    cell_centers_file: str | None = None,
    field_dict: dict | None = None,
) -> tuple[np.ndarray | float, dict]:
    """
    Get indices of cells along a slice given by a direction and a location

    Parameters
    ----------
    case_folder: str
        Path to case folder
    location: float
        Axial location where to pick the cells
        If outside mesh bounds, will raise an error
    direction :  int
        Direction along which to calculate the superficial velocity.
        Must be in [0, 1, 2].
        If None, assume y direction
    tolerance : float
        Include cells where location is in [location - tolerance , location + tolerance].
        If None, it will be 2 times the axial mesh size
    cell_centers_file : str | None
        Filename of cell center data
        If None, finds cell center file automatically
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    ind_location : np.ndarray
        indices of cells along the desired slice
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    if not (f"ind_location_{location:.2g}" in field_dict):

        cell_centers, field_dict = read_cell_centers(
            case_folder=case_folder,
            cell_centers_file=cell_centers_file,
            field_dict=field_dict,
        )

        if direction is None:
            logger.warning(
                "Assuming that axial direction is along the y direction"
            )
            direction = 1

        assert direction in [0, 1, 2]

        axial_cell_centers = np.sort(np.unique(cell_centers[:, direction]))
        if (
            location < axial_cell_centers.min()
            or location > axial_cell_centers.max()
        ):
            raise ValueError(
                f"Location {location:.2g} outside the mesh [{axial_cell_centers.min()}, {axial_cell_centers.max()}]"
            )

        if tolerance is None:
            ind_location_unique = np.argmin(abs(axial_cell_centers - location))
            if ind_location_unique == 0:
                tolerance = 2 * (axial_cell_centers[1] - axial_cell_centers[0])
            elif ind_location_unique == len(axial_cell_centers) - 1:
                tolerance = 2 * (
                    axial_cell_centers[-1] - axial_cell_centers[-2]
                )
            else:
                tolerance = (
                    axial_cell_centers[ind_location_unique + 1]
                    - axial_cell_centers[ind_location_unique - 1]
                )
            logger.debug(
                f"Tolerance for slice location not set, assuming {tolerance:.2g}"
            )

        # Do the actual filtering
        ind_location = np.argwhere(
            abs(cell_centers[:, direction] - location) <= tolerance
        )

        n_cells_location = len(ind_location)

        if n_cells_location == 0:
            raise ValueError(
                f"No cell found for location {location:.2g}, increase tolerance or check if location {location:.2g} is valid"
            )

        logger.debug(
            f"Found {n_cells_location} cells around location {location:.2g}"
        )
        field_dict[f"ind_location_{location:.2g}"] = ind_location

    else:
        ind_location = field_dict[f"ind_location_{location:.2g}"]

    return ind_location, field_dict


def compute_gas_holdup(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str | None = None,
    field_dict: dict | None = None,
) -> tuple[float, dict]:
    r"""
    Calculate volume averaged gas hold up at a given time

    .. math::
       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} (1-\alpha_{\rm liq}) dV

    where:
      - :math:`V_{\rm liq, tot}` is the total volume of liquid in :math:`m^3`
      - :math:`\alpha_{\rm liq}` is the liquid phase volume fraction
      - :math:`V` is the volume of the cells where :math:`\alpha_{\rm liq}` is measured in :math:`m^3`

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    gas_holdup: float
        Volume averaged gas holdup
    field_dict : dict
        Dictionary of fields read
    """

    if field_dict is None:
        field_dict = {}

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }

    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }

    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    ind_liq, field_dict = _get_ind_liq(field_dict=field_dict, **kwargs)
    cell_volume, field_dict = read_cell_volumes(
        field_dict=field_dict, **kwargs_vol
    )

    # Only compute over the pure liquid
    alpha_liq = _field_filter(alpha_liq, ind=ind_liq, field_type="scalar")
    cell_volume = _field_filter(cell_volume, ind=ind_liq, field_type="scalar")

    # Calculate
    gas_holdup = np.sum((1 - alpha_liq) * cell_volume) / np.sum(cell_volume)

    return gas_holdup, field_dict


def compute_superficial_gas_velocity(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str | None = None,
    direction: int | None = None,
    cell_centers_file: str | None = None,
    height: float | None = None,
    use_pv: bool = False,
    field_dict: dict | None = None,
) -> tuple[float, dict]:
    r"""
    Calculate superficial gas velocity (in m/s) in a given direction at a given time

    Without the paraview operations (`use_pv==False`)

    .. math::
       \frac{1}{V_{\rm height, tot}} \int_{V_{\rm height}}  U_{\rm gas} \alpha_{\rm gas} dV

    where:
      - :math:`V_{\rm height, tot}` is the total volume of cells near the axial location considered in :math:`m^3`
      - :math:`\alpha_{\rm gas}` is the gas phase volume fraction
      - :math:`U_{\rm gas}` is the gas phase velocity along the axial direction in :math:`m.s^{-1}`
      - :math:`V_{\rm height}` is the local volume of the cells where :math:`U_{\rm gas} \alpha_{\rm gas}` is measured (near the axial location considered) in :math:`m^3`


    With the paraview operations (`use_pv==True`)

    .. math::
       \frac{1}{S_{\rm height, tot}} \int_{S_{\rm height}}  U_{\rm gas} \alpha_{\rm gas} dS

    where:
      - :math:`S_{\rm height, tot}` is the total area of the slice at the axial location considered and normal tot the direction considered in :math:`m^2`
      - :math:`\alpha_{\rm gas}` is the gas phase volume fraction
      - :math:`U_{\rm gas}` is the gas phase velocity along the axial direction in :math:`m.s^{-1}`
      - :math:`S_{\rm height}` is the local area of the slice where :math:`U_{\rm gas} \alpha_{\rm gas}` is measured (near the axial location considered) in :math:`m^2`


    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    direction :  int | None
        Direction along which to calculate the superficial velocity.
        If None, assume y direction
    cell_centers_file : str | None
        Filename of cell center data
        If None, finds cell center file automatically
    height: float | None
        Axial location at which to compute the superficial velocity.
        If None, use the mid point of the liquid domain along the axial direction
    use_pv: bool
        Use paraview to create a slice in the middle of the reactor
        Default to False
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    sup_vel: float
        Superficial velocity (in m/s)
    field_dict : dict
        Dictionary of fields read
    """

    if field_dict is None:
        field_dict = {}

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    if direction is None:
        logger.warning(
            "Assuming that superficial velocity is along the y direction"
        )
        direction = 1

    if direction not in [0, 1, 2]:
        raise ValueError(f"Direction ({direction}) must be in [0, 1, 2]")

    if use_pv:
        try:
            assert os.path.isdir(
                os.path.join(case_folder, "constant", "polyMesh")
            )
            assert os.path.isfile(
                os.path.join(case_folder, "constant", "polyMesh", "points")
            )
            assert os.path.isfile(
                os.path.join(case_folder, "constant", "polyMesh", "faces")
            )
            assert os.path.isfile(
                os.path.join(case_folder, "constant", "polyMesh", "owner")
            )
            assert os.path.isfile(
                os.path.join(case_folder, "constant", "polyMesh", "neighbour")
            )
            assert os.path.isfile(
                os.path.join(case_folder, "constant", "polyMesh", "boundary")
            )
        except AssertionError:
            logger.warning(
                "Using ParaView requires to make a complete polyMesh, will not use ParaView"
            )
            use_pv = False

    if use_pv:
        logger.debug("Using paraview for superficial velocity calculation")
        # Clean paraview pipeline
        for f in pv.GetSources().values():
            pv.Delete(f)

        # Set up openfoam case
        ofreader = pv.OpenFOAMReader(FileName=case_folder)
        ofreader.CaseType = "Reconstructed Case"
        t = np.array(ofreader.TimestepValues)
        assert t.size > 0

        # Find the time to process
        time_pv_ind = np.argmin(abs(t - float(time_folder)))
        assert abs(t[time_pv_ind] - float(time_folder)) < 1e-12
        pv.UpdatePipeline(time=t[time_pv_ind])

        # Get liquid phase field
        logger.warning("Assuming that alpha_liq > 0.5 denotes pure liquid")
        liquidthreshold = pv.Threshold(
            Input=ofreader,
            Scalars=["CELLS", "alpha.liquid"],
            LowerThreshold=0.5,
            UpperThreshold=1.0,
            ThresholdMethod="Between",
        )

        # Find extent of the liquid phase
        ofvtkdata = pv.servermanager.Fetch(liquidthreshold)
        ofdata = dsa.WrapDataObject(ofvtkdata)
        ofpts = np.array(ofdata.Points.Arrays[0])
        ptsmin_lt = ofpts.min(axis=0)  # minimum values of the three axes
        ptsmax_lt = ofpts.max(axis=0)  # maximum values of the three axes

        # Compute gas velocity in the liquid phase
        if direction == 0:
            u_gas_str = "U.gas_X"
        elif direction == 1:
            u_gas_str = "U.gas_Y"
        elif direction == 2:
            u_gas_str = "U.gas_Z"

        pv_calc = pv.Calculator(
            Input=ofreader,
            AttributeType="Cell Data",
            ResultArrayName="vflowrate",
            Function=f'"alpha.gas"*"{u_gas_str}"',
        )

        # create a new slice in the middle of the liquid domain
        if height is None:
            slice_location = 0.5 * (
                ptsmax_lt[direction] + ptsmin_lt[direction]
            )
        else:
            slice_location = height

        pv_slice = pv.Slice(Input=pv_calc)

        origin = [0.0] * 3
        normal = [0.0] * 3
        origin[direction] = slice_location
        normal[direction] = 1.0

        pv_slice.SliceType.Origin = origin
        pv_slice.SliceType.Normal = normal

        # integrate variables along the slice
        pv_int = pv.IntegrateVariables(Input=pv_slice)

        # calculate superficial vel
        pv.UpdatePipeline(time=t[time_pv_ind])
        pv_dat = dsa.WrapDataObject(pv.servermanager.Fetch(pv_int))
        vfrate = pv_dat.CellData["vflowrate"].item()
        area = pv_dat.CellData["Area"].item()

        sup_vel = vfrate / area

    else:
        kwargs_vol = {
            "case_folder": case_folder,
            "time_folder": volume_time,
            "n_cells": n_cells,
        }
        alpha_gas, field_dict = read_field(
            field_name="alpha.gas", field_dict=field_dict, **kwargs
        )
        U_gas, field_dict = read_field(
            field_name="U.gas", field_dict=field_dict, **kwargs
        )

        if U_gas.shape == (3,):
            # Uniform field
            U_gas_axial = U_gas[direction]
        else:
            # Non-uniform field
            U_gas_axial = U_gas[:, direction]

        cell_volume, field_dict = read_cell_volumes(
            field_dict=field_dict, **kwargs_vol
        )
        cell_centers, field_dict = read_cell_centers(
            case_folder=case_folder,
            cell_centers_file=cell_centers_file,
            field_dict=field_dict,
        )

        if height is None:
            # Find all cells in the middle of the liquid domain
            ind_liq, field_dict = _get_ind_liq(field_dict=field_dict, **kwargs)
            max_dir = np.amax(cell_centers[ind_liq, direction])
            min_dir = np.amin(cell_centers[ind_liq, direction])
            height = (max_dir + min_dir) / 2

        ind_middle, field_dict = _get_ind_slice(
            case_folder=case_folder,
            location=height,
            direction=direction,
            field_dict=field_dict,
        )

        # Filter fields to the right location
        alpha_gas = _field_filter(
            alpha_gas, ind=ind_middle, field_type="scalar"
        )
        cell_volume = _field_filter(
            cell_volume, ind=ind_middle, field_type="scalar"
        )
        U_gas_axial = _field_filter(
            U_gas_axial, ind=ind_middle, field_type="scalar"
        )

        # Compute
        sup_vel = np.sum(U_gas_axial * alpha_gas * cell_volume) / np.sum(
            cell_volume
        )

    return sup_vel, field_dict


def compute_ave_y_liq(
    case_folder: str,
    time_folder: str,
    species_name: str = "CO2",
    n_cells: int | None = None,
    volume_time: str | None = None,
    field_dict: dict | None = None,
) -> tuple[float, dict]:
    r"""
    Calculate liquid volume averaged mass fraction of a species at a given time

    .. math::
       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} Y dV_{\rm liq}

    where:
      - :math:`V_{\rm liq, tot}` is the toal volume of liquid
      - :math:`Y` is the species mass fraction
      - :math:`V_{\rm liq}` is the volume of liquid where :math:`Y` is measured


    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    species_name : str
        Name of the species
    field_dict : dict | None
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    liq_ave_y: float
        Liquid volume averaged mass fraction
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    y_liq, field_dict = read_field(
        field_name=f"{species_name}.liquid", field_dict=field_dict, **kwargs
    )
    ind_liq, field_dict = _get_ind_liq(field_dict=field_dict, **kwargs)

    cell_volume, field_dict = read_cell_volumes(
        field_dict=field_dict, **kwargs_vol
    )

    # Only compute over the liquid
    alpha_liq = _field_filter(alpha_liq, ind=ind_liq, field_type="scalar")
    cell_volume = _field_filter(cell_volume, ind=ind_liq, field_type="scalar")
    y_liq = _field_filter(y_liq, ind=ind_liq, field_type="scalar")

    # Calculate
    liq_ave_y = np.sum(alpha_liq * y_liq * cell_volume) / np.sum(
        alpha_liq * cell_volume
    )

    return liq_ave_y, field_dict


def compute_ave_conc_liq(
    case_folder: str,
    time_folder: str,
    species_name: str = "CO2",
    n_cells: int | None = None,
    volume_time: str | None = None,
    field_dict: dict | None = None,
) -> tuple[float, dict]:
    r"""
    Calculate liquid volume averaged concentration of a species at a given time

    .. math::
       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} \rho_{\rm liq} Y / W dV_{\rm liq}

    where:
      - :math:`V_{\rm liq, tot}` is the toal volume of liquid
      - :math:`\rho_{\rm liq}` is the liquid density
      - :math:`Y` is the species mass fraction
      - :math:`W` is the species molar mass
      - :math:`V_{\rm liq}` is the volume of liquid where :math:`Y` is measured

    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    species_name : str
        Name of the species
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    conc_ave: float
        Liquid volume averaged species concentration
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    mol_weight = species_name_to_mw(
        case_folder=case_folder, species_name=species_name
    )
    logger.debug(
        f"Computing concentration for {species_name} with molecular weight {mol_weight:.4g} kg/mol"
    )

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    y_liq, field_dict = read_field(
        field_name=f"{species_name}.liquid", field_dict=field_dict, **kwargs
    )
    ind_liq, field_dict = _get_ind_liq(field_dict=field_dict, **kwargs)

    cell_volume, field_dict = read_cell_volumes(
        field_dict=field_dict, **kwargs_vol
    )
    try:
        rho_liq, field_dict = read_field(
            field_name="thermo:rho.liquid", field_dict=field_dict, **kwargs
        )
    except FileNotFoundError:
        abs_time_path = os.path.join(case_folder, time_folder)
        logger.warning(
            f"thermo:rho.liquid not found in {abs_time_path}, assuming it is 1000kg/m3"
        )
        rho_liq = 1000.0
        field_dict["rho_liq"] = rho_liq

    # Only compute over the liquid
    alpha_liq = _field_filter(alpha_liq, ind=ind_liq, field_type="scalar")
    cell_volume = _field_filter(cell_volume, ind=ind_liq, field_type="scalar")
    y_liq = _field_filter(y_liq, ind=ind_liq, field_type="scalar")
    rho_liq = _field_filter(rho_liq, ind=ind_liq, field_type="scalar")

    conc_loc = rho_liq * y_liq / mol_weight

    conc_ave = np.sum(conc_loc * alpha_liq * cell_volume) / np.sum(
        alpha_liq * cell_volume
    )

    return conc_ave, field_dict


def compute_ave_bubble_diam(
    case_folder: str,
    time_folder: str,
    n_cells: int | None = None,
    volume_time: str | None = None,
    field_dict: dict | None = None,
) -> tuple[float, dict]:
    r"""
    Calculate averaged bubble diameter over the liquid volume

    .. math::

       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} d_{\rm gas} dV

    where:
      - :math:`V_{\rm liq, tot}` is the toal volume of liquid in :math:`m^3`
      - :math:`d_{\rm gas}` is the bubble diameter in :math:`m`
      - :math:`V_{\rm liq}` is the volume of liquid where :math:`d_{\rm gas}` is measured in :math:`m^3`


    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    diam: float
        Volume averaged gas holdup
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }
    alpha_liq, field_dict = read_field(
        field_name="alpha.liquid", field_dict=field_dict, **kwargs
    )
    d_gas, field_dict = read_bubble_diameter(field_dict=field_dict, **kwargs)
    ind_liq, field_dict = _get_ind_liq(field_dict=field_dict, **kwargs)

    cell_volume, field_dict = read_cell_volumes(
        field_dict=field_dict, **kwargs_vol
    )

    # Only compute over the liquid
    alpha_liq = _field_filter(alpha_liq, ind=ind_liq, field_type="scalar")
    cell_volume = _field_filter(cell_volume, ind=ind_liq, field_type="scalar")
    d_gas = _field_filter(d_gas, ind=ind_liq, field_type="scalar")

    # Calculate
    diam = np.sum(d_gas * alpha_liq * cell_volume) / np.sum(
        alpha_liq * cell_volume
    )

    return diam, field_dict


def compute_instantaneous_kla(
    case_folder: str,
    time_folder: str,
    species_names: str | list[str],
    n_cells: int | None = None,
    volume_time: str | None = None,
    field_dict: dict | None = None,
) -> tuple[dict, dict, dict]:
    r"""
    Calculate :math:`kLa_{\rm spec}` and saturation concentration (:math:`C^*_{\rm spec}`) for a list of species from instantaneous data (rather than doing a fit over time).

    :math:`kLa_{\rm spec}` for the species computed from Eq 7 and 8 in "Computational fluid dynamics study of full-scale aerobic bioreactors: Evaluation of gas–liquid mass transfer, oxygen uptake, and dynamic oxygen distribution", M. J. Rahimi, H. Sitaraman, D. Humbird, J. J. Stickel, Chem. Eng. Research and Design, Vol. 139, pp 293-295, 2018.



    .. math::

       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} kLa_{\rm spec} dV

    .. math::

       kLa_{\rm spec} = 3600 \sqrt{\frac{4 D_{\rm spec} |u_{\rm slip}|}{\pi d_{\rm gas}}} \frac{6 \alpha_{\rm gas}}{d_{\rm gas}}

    .. math::

       kLa_{\rm spec} = (\frac{2}{\pi^{1/2}} \times 3600) Re^{1/2} \frac{\mu_{\rm liq}^{1/2}}{D_{\rm spec}^{1/2} \rho_{\rm liq}^{1/2}} \frac{D_{\rm spec}}{d_{\rm gas}} \frac{6}{d_{\rm gas}} \alpha_{\rm gas}

    .. math::

       Re = \frac{\rho_{\rm liq} |u_{\rm slip}| d_{\rm gas}}{\mu_{\rm liq}}

    where:
      - :math:`kLa_{\rm spec}` is the mass transfer rate in :math:`h^{-1}`
      - :math:`d_{\rm gas}` is the bubble diameter in :math:`m`
      - :math:`\alpha_{\rm gas}` is the volume fraction of gas
      - :math:`\mu_{\rm liq}` is the liquid viscosity in :math:`kg.m^{-1}.s^{-1}`
      - :math:`\rho_{\rm liq}` is the liquid density in :math:`kg.m^{-3}`
      - :math:`D_{\rm spec}` is the species molecular diffusivity in :math:`m^2.s^{-1}`
      - :math:`|u_{\rm slip}|` is the magnitude of the slip velocity in :math:`m.s^{-1}`
      - :math:`V_{\rm liq}` is the volume of liquid in :math:`m^3`

     .. math::

       \frac{1}{V_{\rm liq, tot}} \int_{V_{\rm liq}} C^*_{\rm spec} dV

    :math:`C^*_{\rm spec}` computed from Eq 10 in "Computational fluid dynamics study of full-scale aerobic bioreactors: Evaluation of gas–liquid mass transfer, oxygen uptake, and dynamic oxygen distribution", M. J. Rahimi, H. Sitaraman, D. Humbird, J. J. Stickel, Chem. Eng. Research and Design, Vol. 139, pp 293-295, 2018.

     .. math::

       C^*_{\rm spec} = \rho_{\rm gas} Y_{\rm spec, gas} He_{\rm spec} / W_{\rm spec}

     and
      - :math:`C^{*}_{\rm spec}` is the saturation concentration of species spec in :math:`mol.m^{-3}`
      - :math:`\rho_{\rm gas}` is the density of the gas in :math:`kg.m^{-3}`
      - :math:`Y_{\rm spec, gas}` is the mass fraction of species spec in the gas phase
      - :math:`He_{\rm spec}` is the Henry's constant of species spec
      - :math:`W_{\rm spec}` is the molar mass of species spec in :math:`kg.mol^{-1}`


    Parameters
    ----------
    case_folder: str
        Path to case folder
    time_folder: str
        Name of time folder to analyze
    species_names: str | list[str]
        List of species name for which to compute kla
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    kla_spec: dict
        Instantaneous volume averaged kLa for each species
        Keys are species names
        Values are the kLa values
    cstar_spec: dict
        Instantaneous volume averaged cstar for each species
        Keys are species names
        Values are the cstar values
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    if isinstance(species_names, str):
        species_names = [species_names]

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "time_folder": time_folder,
        "n_cells": n_cells,
    }
    kwargs_vol = {
        "case_folder": case_folder,
        "time_folder": volume_time,
        "n_cells": n_cells,
    }

    # Read globarVars into a python dict
    # Replace all the #calc entries with their numeral values
    globalVars = read_global_vars(case_folder=case_folder, cross_ref=True)

    # Check that global vars has the values we want and provide a useful error message otherwise
    mw_species = {}
    for species_name in species_names:
        if not f"He_{species_name}" in globalVars:
            err_msg = f"He_{species_name} was not found in globalVars."
            err_msg += f'\nIf you add it, it should be looking like #calc "$H_{species_name}_298 * exp($DH_{species_name} *(1. / $T0 - 1./298.15))";'
            raise KeyError(err_msg)
        if not f"D_{species_name}" in globalVars:
            err_msg = f"D_{species_name} was not found in globalVars."
            err_msg += f'\nIf you add it, it should be looking like #calc "1.173e-16 * pow($WC_psi * $WC_M,0.5) * $T0 / $muMixLiq / pow($WC_V_{species_name},0.6)";'
            raise KeyError(err_msg)
        mw_species[species_name] = species_name_to_mw(
            case_folder=case_folder, species_name=species_name
        )

    # Get liquid domain
    ind_liq, field_dict = _get_ind_liq(field_dict=field_dict, **kwargs)

    # Read all the fields
    alpha_gas, field_dict = read_field(
        field_name="alpha.gas", field_dict=field_dict, **kwargs
    )
    rho_liq, field_dict = read_field(
        field_name="thermo:rho.liquid", field_dict=field_dict, **kwargs
    )
    rho_gas, field_dict = read_field(
        field_name="thermo:rho.gas", field_dict=field_dict, **kwargs
    )
    U_gas, field_dict = read_field(
        field_name="U.gas", field_dict=field_dict, **kwargs
    )
    U_liq, field_dict = read_field(
        field_name="U.liquid", field_dict=field_dict, **kwargs
    )
    d_gas, field_dict = read_bubble_diameter(field_dict=field_dict, **kwargs)
    mu_liq, field_dict = read_mu_liquid(field_dict=field_dict, **kwargs)
    species_gas = {}
    for species_name in species_names:
        species_gas[species_name], field_dict = read_field(
            field_name=f"{species_name}.gas", field_dict=field_dict, **kwargs
        )

    # Only compute over the liquid
    alpha_gas = _field_filter(alpha_gas, ind=ind_liq, field_type="scalar")
    alpha_liq = 1 - alpha_gas
    rho_liq = _field_filter(rho_liq, ind=ind_liq, field_type="scalar")
    rho_gas = _field_filter(rho_gas, ind=ind_liq, field_type="scalar")
    U_gas = _field_filter(U_gas, ind=ind_liq, field_type="vector")
    U_liq = _field_filter(U_liq, ind=ind_liq, field_type="vector")
    d_gas = _field_filter(d_gas, ind=ind_liq, field_type="scalar")
    mu_liq = _field_filter(mu_liq, ind=ind_liq, field_type="scalar")
    for species_name in species_names:
        species_gas[species_name] = _field_filter(
            species_gas[species_name], ind=ind_liq, field_type="scalar"
        )

    mag_U_diff = np.sqrt(
        (U_gas[:, 0] - U_liq[:, 0]) ** 2
        + (U_gas[:, 1] - U_liq[:, 1]) ** 2
        + (U_gas[:, 2] - U_liq[:, 2]) ** 2
    )

    # Compute kLa
    Re = rho_liq * mag_U_diff * d_gas / mu_liq
    kla_spec_field = {}
    for species_name in species_names:
        kla_spec_field[species_name] = (
            (2 / np.pi**0.5)
            * 3600
            * (Re**0.5)
            * (((mu_liq / rho_liq) / globalVars[f"D_{species_name}"]) ** 0.5)
            * (globalVars[f"D_{species_name}"] / d_gas)
            * (6.0 / d_gas)
            * alpha_gas
        )
    cstar_spec_field = {}
    for species_name in species_names:
        cstar_spec_field[species_name] = (
            rho_gas
            * species_gas[species_name]
            * globalVars[f"He_{species_name}"]
        ) / mw_species[species_name]

    # Do volume average
    cell_volume, field_dict = read_cell_volumes(
        field_dict=field_dict, **kwargs_vol
    )
    cell_volume = _field_filter(cell_volume, ind=ind_liq, field_type="scalar")

    kla_spec = {}
    cstar_spec = {}
    for species_name in species_names:
        kla_spec[species_name] = np.sum(
            cell_volume * kla_spec_field[species_name]
        ) / np.sum(cell_volume)
        cstar_spec[species_name] = np.sum(
            cell_volume * alpha_liq * cstar_spec_field[species_name]
        ) / np.sum(cell_volume * alpha_liq)

    return kla_spec, cstar_spec, field_dict


def compute_fitted_kla(
    case_folder: str,
    species_names: str | list[str],
    n_cells: int | None = None,
    volume_time: str | None = None,
    num_warmup: int = 4000,
    num_samples: int = 1000,
    field_dict: dict | None = None,
) -> tuple[dict, dict, dict]:
    r"""
    Calculate :math:`kLa_{\rm spec}` and saturation concentration (:math:`C^*_{\rm spec}`) for a list of species from time series data (rather than instantaneously).

    Given a time series of concentration of species, the following expression is fitted

    .. math::
       [spec](t) =  [spec]^* (1 - \operatorname{exp}(-{kLa}_{\rm spec} t)).

    where

      - :math:`kLa_{\rm spec}` is the mass transfer rate of species :math:`\rm spec` in :math:`h^{-1}`
      - :math:`t` is the time in :math:`s`
      - :math:`[spec]^*` is the estimated saturation concentration of species :math:`\rm spec` in :math:`mol/m^3`
      - :math:`[spec](t)` is the instantaneous liquid volume averaged concentration of species :math:`\rm spec` in :math:`mol/m^3`

    Both :math:`[spec]^*` and :math:`kLa_{\rm spec}` are fitted.
    The fit is done with Markov Chain Monte Carlo which outputs samples of the posterior PDF of :math:`[spec]^*` and :math:`kLa_{\rm spec}`.

    Parameters
    ----------
    case_folder: str
        Path to case folder
    species_names: str | list[str]
        List of species name for which to compute kla
    n_cells : int | None
        Number of cells in the domain.
        If None, it will deduced from the field reading
    volume_time : str | None
        Time folder to read to get the cell volumes.
        If None, finds volume time automatically
    num_warmup: int
        Number of MCMC samples in the warmup phase
        Defaults to 4000
    num_samples: int
        Number of posterior MCMC samples generated
        Defaults to 1000
    field_dict : dict
        Dictionary of fields used to avoid rereading the same fields to calculate different quantities

    Returns
    ----------
    kla_spec: dict
        Instantaneous volume averaged kLa for each species
        Keys are species names
        Values are dictionaries with key 'mean' (mean kLa value) and 'std' (1 standard deviation for the kLa value)
    cstar_spec: dict
        Instantaneous volume averaged cstar for each species
        Keys are species names
        Values are dictionaries with key 'mean' (mean cstar value) and 'std' (1 standard deviation for the cstar value)
    field_dict : dict
        Dictionary of fields read
    """
    if field_dict is None:
        field_dict = {}

    if isinstance(species_names, str):
        species_names = [species_names]

    # Read relevant fields
    kwargs = {
        "case_folder": case_folder,
        "n_cells": n_cells,
        "volume_time": volume_time,
    }

    # Get all the time folders
    time_float_sorted, time_str_sorted = get_case_times(case_folder)

    # Read globarVars into a python dict
    # Replace all the #calc entries with their numeral values
    globalVars = read_global_vars(case_folder=case_folder, cross_ref=True)

    # Get Mw of the species
    mw_species = {}
    for species_name in species_names:
        mw_species[species_name] = species_name_to_mw(
            case_folder=case_folder, species_name=species_name
        )

    # Initialize the mesh fields
    mesh_field_dict = {}
    if "cell_centers" in field_dict:
        mesh_field_dict["cell_centers"] = field_dict["cell_centers"]
    else:
        mesh_field_dict["cell_centers"], _ = read_cell_centers(case_folder)
    if "V" in field_dict:
        mesh_field_dict["V"] = field_dict["V"]
    else:
        mesh_field_dict["V"], _ = read_cell_volumes(case_folder)

    logger.info("Reading the species concentration history")

    # Initialize the data structure for concentration
    c_history = {}
    for species_name in species_names:
        c_history[species_name] = np.zeros(len(time_str_sorted))

    # Read concentration
    for itime, time_folder in enumerate(time_str_sorted):
        logger.debug(f"Reading {time_folder}")
        # Reinitialize kla field dict
        kla_field_dict = {}
        for key in mesh_field_dict:
            kla_field_dict[key] = mesh_field_dict[key]

        # Compute reactor averaged liquid concentration for all the species
        for species_name in species_names:
            c_liq, kla_field_dict = compute_ave_conc_liq(
                time_folder=time_folder,
                species_name=species_name,
                field_dict=kla_field_dict,
                **kwargs,
            )
        c_history[species_name][itime] = c_liq

    logger.info("Doing kla fit")
    # Compute kla
    kla_spec = {}
    cstar_spec = {}
    for species_name in species_names:
        kla_res = compute_kla(
            np.array(time_float_sorted),
            c_history[species_name],
            num_warmup=num_warmup,
            num_samples=num_samples,
        )
        # Convert to h-1
        kla_spec[species_name] = {
            "mean": kla_res["kla"] * 3600,
            "std": kla_res["kla_err"] * 3600,
        }
        cstar_spec[species_name] = {
            "mean": kla_res["cstar"],
            "std": kla_res["cstar_err"],
        }

    return kla_spec, cstar_spec, field_dict



import json


def generate_stl_patch(filename, bc_dict, geom_dict):
    final_dict = {}
    final_dict["Geometry"] = geom_dict["Geometry"]
    for patch in bc_dict:
        final_dict[patch] = bc_dict[patch]
    with open(filename, "w+") as f:
        json.dump(final_dict, f, indent=2)


def generate_dynamic_mixer(filename, mixers_list, geom_dict):
    final_dict = {}
    final_dict["Meshing"] = geom_dict["Meshing"]
    final_dict["Geometry"] = geom_dict["Geometry"]
    final_dict["mixers"] = mixers_list
    with open(filename, "w+") as f:
        json.dump(final_dict, f, indent=2)


def make_default_geom_dict_from_file(filename, rescale=2.7615275385627096):
    with open(filename, "r+") as f:
        geom_dict = json.load(f)
    if "rescale" not in geom_dict["Geometry"]["OverallDomain"]["x"]:
        geom_dict["Geometry"]["OverallDomain"]["x"]["rescale"] = rescale
        geom_dict["Geometry"]["OverallDomain"]["y"]["rescale"] = rescale
        geom_dict["Geometry"]["OverallDomain"]["z"]["rescale"] = rescale
    assert "Meshing" in geom_dict
    assert "Geometry" in geom_dict
    return geom_dict



import logging
from pathlib import Path


class ConditionalFormatter(logging.Formatter):
    def format(self, record):
        """Change how to log if it is an INFO or a WARNING/ERROR"""
        if record.levelno >= logging.WARNING:
            self._style._fmt = "%(asctime)s [%(levelname)s] bird: %(message)s"
        else:
            self._style._fmt = "%(asctime)s [%(levelname)s] bird: %(message)s"
        return super().format(record)


def setup_logging(logfile: str | None = None, level=logging.INFO):
    """Setup logging"""
    logger = logging.getLogger()
    logger.setLevel(level)

    formatter = ConditionalFormatter()

    # Console handler
    ch = logging.StreamHandler()
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # Optional file handler
    if logfile:
        fh = logging.FileHandler(Path(logfile))
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    return logger
